from __future__ import annotations

import logging
from typing import (
    List, Optional, Sequence, TYPE_CHECKING, Union)
from opentrons.broker import Broker
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons import types, hardware_control as hc
from opentrons.commands import commands as cmds
from opentrons.commands.publisher import CommandPublisher, do_publish, publish
from opentrons.hardware_control.types import PipettePair
from opentrons.protocols.advanced_control.mix import mix_from_kwargs
from opentrons.protocols.api_support.instrument import \
    validate_blowout_location, tip_length_for, validate_tiprack, \
    determine_drop_target, validate_can_aspirate, validate_can_dispense
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocol_api.module_contexts import ThermocyclerContext
from opentrons.protocols.api_support.util import (
    FlowRates, PlungerSpeeds, Clearances, clamp_value, requires_version)
from opentrons.protocols.context.instrument import \
    AbstractInstrument
from opentrons.protocols.api_support.types import APIVersion
from .labware import (
    Labware, OutOfTipsError, Well, next_available_tip)
from opentrons.protocols.advanced_control import transfers
from .paired_instrument_context import (
    PairedInstrumentContext, UnsupportedInstrumentPairingError)

if TYPE_CHECKING:
    from opentrons.protocol_api import ProtocolContext

AdvancedLiquidHandling = Union[
    Well,
    types.Location,
    List[Union[Well, types.Location]],
    List[List[Well]]]


class InstrumentContext(CommandPublisher):
    """ A context for a specific pipette or instrument.

    This can be used to call methods related to pipettes - moves or
    aspirates or dispenses, or higher-level methods.

    Instances of this class bundle up state and config changes to a
    pipette - for instance, changes to flow rates or trash containers.
    Action methods (like :py:meth:`aspirate` or :py:meth:`distribute`) are
    defined here for convenience.

    In general, this class should not be instantiated directly; rather,
    instances are returned from :py:meth:`ProtocolContext.load_instrument`.

    .. versionadded:: 2.0

    """

    def __init__(self,
                 implementation: AbstractInstrument,
                 ctx: ProtocolContext,
                 broker: Broker,
                 log_parent: logging.Logger,
                 at_version: APIVersion,
                 tip_racks: List[Labware] = None,
                 trash: Optional[Labware] = None,
                 ) -> None:

        super().__init__(broker)
        self._api_version = at_version
        self._implementation = implementation
        self._ctx = ctx
        self._log = log_parent.getChild(repr(self))

        self._tip_racks = tip_racks or list()
        for tip_rack in self.tip_racks:
            assert tip_rack.is_tiprack
            validate_tiprack(self.name, tip_rack, self._log)
        if trash is None:
            self.trash_container = self._ctx.fixed_trash
        else:
            self.trash_container = trash

        self._last_tip_picked_up_from: Union[Well, None] = None
        self._starting_tip: Union[Well, None] = None
        self.requested_as = self._implementation.get_instrument_name()

    @property  # type: ignore
    @requires_version(2, 0)
    def api_version(self) -> APIVersion:
        return self._api_version

    @property  # type: ignore
    @requires_version(2, 0)
    def starting_tip(self) -> Union[Well, None]:
        """ The starting tip from which the pipette pick up
        """
        return self._starting_tip

    @starting_tip.setter
    def starting_tip(self, location: Union[Well, None]):
        self._starting_tip = location

    @requires_version(2, 0)
    def reset_tipracks(self):
        """ Reload all tips in each tip rack and reset starting tip
        """
        for tiprack in self.tip_racks:
            tiprack.reset()
        self.starting_tip = None

    @property  # type: ignore
    @requires_version(2, 0)
    def default_speed(self) -> float:
        """ The speed at which the robot's gantry moves.

        By default, 400 mm/s. Changing this value will change the speed of the
        pipette when moving between labware. In addition to changing the
        default, the speed of individual motions can be changed with the
        ``speed`` argument to :py:meth:`InstrumentContext.move_to`.
        """
        return self._implementation.get_default_speed()

    @default_speed.setter
    def default_speed(self, speed: float):
        self._implementation.set_default_speed(speed)

    @requires_version(2, 0)  # noqa: C901
    def aspirate(self,
                 volume: Optional[float] = None,
                 location: Union[types.Location, Well] = None,
                 rate: float = 1.0) -> InstrumentContext:
        """
        Aspirate a given volume of liquid from the specified location, using
        this pipette.

        :param volume: The volume to aspirate, in microliters (µL).  If 0 or
                       unspecified, defaults to the highest volume possible
                       with this pipette and its currently attached tip.
        :type volume: int or float
        :param location: Where to aspirate from. If `location` is a
                         :py:class:`.Well`, the robot will aspirate from
                         :py:obj:`well_bottom_clearance.aspirate` mm
                         above the bottom of the well. If `location` is a
                         :py:class:`.Location` (i.e. the result of
                         :py:meth:`.Well.top` or :py:meth:`.Well.bottom`), the
                         robot will aspirate from the exact specified location.
                         If unspecified, the robot will aspirate from the
                         current position.
        :param rate: A relative modifier for how quickly to aspirate liquid.
                     The flow rate for this aspirate will be
                     `rate` * :py:attr:`flow_rate.aspirate <flow_rate>`.
                     If not specified, defaults to 1.0.
        :type rate: float
        :returns: This instance.

        .. note::

            If ``aspirate`` is called with a single argument, it will not try
            to guess whether the argument is a volume or location - it is
            required to be a volume. If you want to call ``aspirate`` with only
            a location, specify it as a keyword argument:
            ``instr.aspirate(location=wellplate['A1'])``

        """
        self._log.debug("aspirate {} from {} at {}"
                        .format(volume,
                                location if location else 'current position',
                                rate))

        if isinstance(location, Well):
            dest = location.bottom().move(
                types.Point(0, 0, self.well_bottom_clearance.aspirate))
        elif isinstance(location, types.Location):
            dest = location
        elif location is not None:
            raise TypeError(
                'location should be a Well or Location, but it is {}'
                .format(location))
        elif self._ctx.location_cache:
            dest = self._ctx.location_cache
        else:
            raise RuntimeError(
                "If aspirate is called without an explicit location, another"
                " method that moves to a location (such as move_to or "
                "dispense) must previously have been called so the robot "
                "knows where it is.")
        if self.api_version >= APIVersion(2, 11):
            validate_can_aspirate(dest)

        if self.current_volume == 0:
            # Make sure we're at the top of the labware and clear of any
            # liquid to prepare the pipette for aspiration

            if self.api_version < APIVersion(2, 3) or \
                    not self._implementation.is_ready_to_aspirate():
                if dest.labware.is_well:
                    self.move_to(dest.labware.as_well().top(),
                                 publish=False)
                else:
                    # TODO(seth,2019/7/29): This should be a warning exposed
                    #  via rpc to the runapp
                    self._log.warning(
                        "When aspirate is called on something other than a "
                        "well relative position, we can't move to the top of"
                        " the well to prepare for aspiration. This might "
                        "cause over aspiration if the previous command is a "
                        "blow_out.")
                self._implementation.prepare_for_aspirate()
            self.move_to(dest, publish=False)
        elif dest != self._ctx.location_cache:
            self.move_to(dest, publish=False)

        c_vol = self._implementation.get_available_volume() \
            if not volume else volume

        do_publish(self.broker, cmds.aspirate, self.aspirate,
                   'before', None, None, self, c_vol, dest, rate)
        self._implementation.aspirate(volume=c_vol, rate=rate)
        do_publish(self.broker, cmds.aspirate, self.aspirate,
                   'after', self, None, self, c_vol, dest, rate)
        return self

    @requires_version(2, 0)
    def dispense(self,
                 volume: Optional[float] = None,
                 location: Union[types.Location, Well] = None,
                 rate: float = 1.0) -> InstrumentContext:
        """
        Dispense a volume of liquid (in microliters/uL) using this pipette
        into the specified location.

        If only a volume is passed, the pipette will dispense from its current
        position. If only a location is passed (as in
        ``instr.dispense(location=wellplate['A1'])``), all of the liquid
        aspirated into the pipette will be dispensed (this volume is accessible
        through :py:attr:`current_volume`).

        :param volume: The volume of liquid to dispense, in microliters. If 0
                       or unspecified, defaults to :py:attr:`current_volume`.
        :type volume: int or float

        :param location: Where to dispense into. If `location` is a
                         :py:class:`.Well`, the robot will dispense into
                         :py:obj:`well_bottom_clearance.dispense` mm
                         above the bottom of the well. If `location` is a
                         :py:class:`.Location` (i.e. the result of
                         :py:meth:`.Well.top` or :py:meth:`.Well.bottom`), the
                         robot will dispense into the exact specified location.
                         If unspecified, the robot will dispense into the
                         current position.
        :param rate: A relative modifier for how quickly to dispense liquid.
                     The flow rate for this dispense will be
                     `rate` * :py:attr:`flow_rate.dispense <flow_rate>`.
                     If not specified, defaults to 1.0.
        :type rate: float

        :returns: This instance.

        .. note::

            If ``dispense`` is called with a single argument, it will not try
            to guess whether the argument is a volume or location - it is
            required to be a volume. If you want to call ``dispense`` with only
            a location, specify it as a keyword argument:
            ``instr.dispense(location=wellplate['A1'])``

        """
        self._log.debug("dispense {} from {} at {}"
                        .format(volume,
                                location if location else 'current position',
                                rate))
        if isinstance(location, Well):
            if LabwareLike(location).is_fixed_trash():
                loc = location.top()
            else:
                loc = location.bottom().move(
                    types.Point(0, 0, self.well_bottom_clearance.dispense))
            self.move_to(loc, publish=False)
        elif isinstance(location, types.Location):
            loc = location
            self.move_to(location, publish=False)
        elif location is not None:
            raise TypeError(
                f'location should be a Well or Location, but it is {location}')
        elif self._ctx.location_cache:
            loc = self._ctx.location_cache
        else:
            raise RuntimeError(
                "If dispense is called without an explicit location, another"
                " method that moves to a location (such as move_to or "
                "aspirate) must previously have been called so the robot "
                "knows where it is.")
        if self.api_version >= APIVersion(2, 11):
            validate_can_dispense(loc)

        c_vol = self.current_volume if not volume else volume

        do_publish(self.broker, cmds.dispense, self.dispense,
                   'before', None, None, self, c_vol, loc, rate)
        self._implementation.dispense(volume=c_vol, rate=rate)
        do_publish(self.broker, cmds.dispense, self.dispense,
                   'after', self, None, self, c_vol, loc, rate)
        return self

    @requires_version(2, 0)
    def mix(self,
            repetitions: int = 1,
            volume: Optional[float] = None,
            location: Union[types.Location, Well] = None,
            rate: float = 1.0) -> InstrumentContext:
        """
        Mix a volume of liquid (uL) using this pipette, by repeatedly
        aspirating and dispensing in the same place.

        :param repetitions: how many times the pipette should mix (default: 1)
        :param volume: number of microliters to mix.  If 0 or unspecified,
                       defaults to the highest volume possible with this
                       pipette and its currently attached tip.
        :param location: a Well or a position relative to well.
                         e.g, `plate.rows()[0][0].bottom()`.  If unspecified,
                         the pipette will mix from its current position.
        :type location: types.Location
        :param rate: A relative modifier for how quickly to aspirate and
                     dispense liquid during this mix. When aspirating, the flow
                     rate will be
                     `rate` * :py:attr:`flow_rate.aspirate <flow_rate>`,
                     and when dispensing, it will be
                     `rate` * :py:attr:`flow_rate.dispense <flow_rate>`.
        :raises NoTipAttachedError: If no tip is attached to the pipette.
        :returns: This instance

        .. note::

            All the arguments to ``mix`` are optional; however, if you do
            not want to specify one of them, all arguments after that one
            should be keyword arguments. For instance, if you do not want
            to specify volume, you would call
            ``pipette.mix(1, location=wellplate['A1'])``. If you do not
            want to specify repetitions, you would call
            ``pipette.mix(volume=10, location=wellplate['A1'])``. Unlike
            previous API versions, ``mix`` will not attempt to guess your
            inputs; the first argument will always be interpreted as
            ``repetitions``, the second as ``volume``, and the third as
            ``location`` unless you use keywords.

        """
        self._log.debug(
            'mixing {}uL with {} repetitions in {} at rate={}'.format(
                volume, repetitions,
                location if location else 'current position', rate))
        if not self._implementation.has_tip():
            raise hc.NoTipAttachedError('Pipette has no tip. Aborting mix()')

        c_vol = self._implementation.get_available_volume() \
            if not volume else volume

        do_publish(self.broker, cmds.mix, self.mix,
                   'before', None, None,
                   self, repetitions, c_vol, location)
        self.aspirate(volume, location, rate)
        while repetitions - 1 > 0:
            self.dispense(volume, rate=rate)
            self.aspirate(volume, rate=rate)
            repetitions -= 1
        self.dispense(volume, rate=rate)
        do_publish(self.broker, cmds.mix, self.mix,
                   'after', None, None,
                   self, repetitions, c_vol, location)
        return self

    @requires_version(2, 0)
    def blow_out(self,
                 location: Union[types.Location, Well] = None
                 ) -> InstrumentContext:
        """
        Blow liquid out of the tip.

        If :py:attr:`dispense` is used to completely empty a pipette,
        usually a small amount of liquid will remain in the tip. This
        method moves the plunger past its usual stops to fully remove
        any remaining liquid from the tip. Regardless of how much liquid
        was in the tip when this function is called, after it is done
        the tip will be empty.

        :param location: The location to blow out into. If not specified,
                         defaults to the current location of the pipette
        :type location: :py:class:`.Well` or :py:class:`.Location` or None

        :raises RuntimeError: If no location is specified and location cache is
                              None. This should happen if `blow_out` is called
                              without first calling a method that takes a
                              location (eg, :py:meth:`.aspirate`,
                              :py:meth:`dispense`)
        :returns: This instance
        """

        if isinstance(location, Well):
            if location.parent.is_tiprack:
                self._log.warning('Blow_out being performed on a tiprack. '
                                  'Please re-check your code')
            loc = location.top()
            self.move_to(loc, publish=False)
        elif isinstance(location, types.Location):
            loc = location
            self.move_to(loc, publish=False)
        elif location is not None:
            raise TypeError(
                'location should be a Well or Location, but it is {}'
                .format(location))
        elif self._ctx.location_cache:
            # if location cache exists, pipette blows out immediately at
            # current location, no movement is needed
            pass
        else:
            raise RuntimeError(
                "If blow out is called without an explicit location, another"
                " method that moves to a location (such as move_to or "
                "dispense) must previously have been called so the robot "
                "knows where it is.")

        do_publish(self.broker, cmds.blow_out, self.blow_out, 'before',
                   None, None, self, location or self._ctx.location_cache)
        self._implementation.blow_out()
        do_publish(self.broker, cmds.blow_out, self.blow_out, 'after',
                   None, None, self, location or self._ctx.location_cache)
        return self

    def _determine_speed(self, speed: float):
        if self.api_version < APIVersion(2, 4):
            return clamp_value(speed, 80, 20, 'touch_tip:')
        else:
            return clamp_value(speed, 80, 1, 'touch_tip:')

    @publish.both(command=cmds.touch_tip)
    @requires_version(2, 0)
    def touch_tip(self,
                  location: Optional[Well] = None,
                  radius: float = 1.0,
                  v_offset: float = -1.0,
                  speed: float = 60.0) -> InstrumentContext:
        """
        Touch the pipette tip to the sides of a well, with the intent of
        removing left-over droplets

        :param location: If no location is passed, pipette will
                         touch tip at current well's edges
        :type location: :py:class:`.Well` or None
        :param radius: Describes the proportion of the target well's
                       radius. When `radius=1.0`, the pipette tip will move to
                       the edge of the target well; when `radius=0.5`, it will
                       move to 50% of the well's radius. Default: 1.0 (100%)
        :type radius: float
        :param v_offset: The offset in mm from the top of the well to touch tip
                         A positive offset moves the tip higher above the well,
                         while a negative offset moves it lower into the well
                         Default: -1.0 mm
        :type v_offset: float
        :param speed: The speed for touch tip motion, in mm/s.
                      Default: 60.0 mm/s, Max: 80.0 mm/s, Min: 20.0 mm/s
        :type speed: float
        :raises NoTipAttachedError: if no tip is attached to the pipette
        :raises RuntimeError: If no location is specified and location cache is
                              None. This should happen if `touch_tip` is called
                              without first calling a method that takes a
                              location (eg, :py:meth:`.aspirate`,
                              :py:meth:`dispense`)
        :returns: This instance

        .. note::

            This is behavior change from legacy API (which accepts any
            :py:class:`.Placeable` as the ``location`` parameter)

        """
        if not self._implementation.has_tip():
            raise hc.NoTipAttachedError('Pipette has no tip to touch_tip()')

        checked_speed = self._determine_speed(speed)

        # If location is a valid well, move to the well first
        if location is None:
            last_location = self._ctx.location_cache
            if not last_location:
                raise RuntimeError('No valid current location cache present')
            else:
                well = last_location.labware
                # type checked below
        else:
            well = LabwareLike(location)

        if well.is_well:
            if 'touchTipDisabled' in well.quirks_from_any_parent():
                self._log.info(f"Ignoring touch tip on labware {well}")
                return self
            if well.parent.as_labware().is_tiprack:
                self._log.warning('Touch_tip being performed on a tiprack. '
                                  'Please re-check your code')

            if self.api_version < APIVersion(2, 4):
                to_loc = well.as_well().top()
            else:
                move_with_z_offset =\
                    well.as_well().top().point + types.Point(0, 0, v_offset)
                to_loc = types.Location(move_with_z_offset, well)
            self.move_to(to_loc, publish=False)
        else:
            raise TypeError(
                'location should be a Well, but it is {}'.format(location))

        self._implementation.touch_tip(
            location=well.as_well()._impl,
            radius=radius,
            v_offset=v_offset,
            speed=checked_speed
        )
        return self

    @publish.both(command=cmds.air_gap)
    @requires_version(2, 0)
    def air_gap(self,
                volume: Optional[float] = None,
                height: Optional[float] = None) -> InstrumentContext:
        """
        Pull air into the pipette current tip at the current location

        :param volume: The amount in uL to aspirate air into the tube.
                       (Default will use all remaining volume in tip)
        :type volume: float

        :param height: The number of millimiters to move above the current Well
                       to air-gap aspirate. (Default: 5mm above current Well)
        :type height: float

        :raises NoTipAttachedError: If no tip is attached to the pipette

        :raises RuntimeError: If location cache is None.
                              This should happen if `touch_tip` is called
                              without first calling a method that takes a
                              location (eg, :py:meth:`.aspirate`,
                              :py:meth:`dispense`)

        :returns: This instance

        .. note::

            Both ``volume`` and height are optional, but unlike previous API
            versions, if you want to specify only ``height`` you must do it
            as a keyword argument: ``pipette.air_gap(height=2)``. If you
            call ``air_gap`` with only one unnamed argument, it will always
            be interpreted as a volume.


        """
        if not self._implementation.has_tip():
            raise hc.NoTipAttachedError('Pipette has no tip. Aborting air_gap')

        if height is None:
            height = 5
        loc = self._ctx.location_cache
        if not loc or not loc.labware.is_well:
            raise RuntimeError('No previous Well cached to perform air gap')
        target = loc.labware.as_well().top(height)
        self.move_to(target, publish=False)
        self.aspirate(volume)
        return self

    @publish.both(command=cmds.return_tip)
    @requires_version(2, 0)
    def return_tip(self, home_after: bool = True) -> InstrumentContext:
        """
        If a tip is currently attached to the pipette, then it will return the
        tip to it's location in the tiprack.

        It will not reset tip tracking so the well flag will remain False.

        :returns: This instance

        :param home_after:
            See the ``home_after`` parameter in :py:obj:`drop_tip`.
        """
        if not self._implementation.has_tip():
            self._log.warning('Pipette has no tip to return')
        loc = self._last_tip_picked_up_from
        if not isinstance(loc, Well):
            raise TypeError('Last tip location should be a Well but it is: '
                            '{}'.format(loc))
        return_height = self._implementation.get_return_height()
        drop_loc = determine_drop_target(self.api_version,
                                         loc,
                                         return_height,
                                         APIVersion(2, 3))
        self.drop_tip(drop_loc, home_after=home_after)

        return self

    @requires_version(2, 0)
    def pick_up_tip(
            self, location: Union[types.Location, Well] = None,
            presses: Optional[int] = None,
            increment: Optional[float] = None) -> InstrumentContext:
        """
        Pick up a tip for the pipette to run liquid-handling commands with

        If no location is passed, the Pipette will pick up the next available
        tip in its :py:attr:`InstrumentContext.tip_racks` list.

        The tip to pick up can be manually specified with the `location`
        argument. The `location` argument can be specified in several ways:

        * If the only thing to specify is which well from which to pick
          up a tip, `location` can be a :py:class:`.Well`. For instance,
          if you have a tip rack in a variable called `tiprack`, you can
          pick up a specific tip from it with
          ``instr.pick_up_tip(tiprack.wells()[0])``. This style of call can
          be used to make the robot pick up a tip from a tip rack that
          was not specified when creating the :py:class:`.InstrumentContext`.

        * If the position to move to in the well needs to be specified,
          for instance to tell the robot to run its pick up tip routine
          starting closer to or farther from the top of the tip,
          `location` can be a :py:class:`.types.Location`; for instance,
          you can call ``instr.pick_up_tip(tiprack.wells()[0].top())``.

        :param location: The location from which to pick up a tip.
        :type location: :py:class:`.types.Location` or :py:class:`.Well` to
                        pick up a tip from.
        :param presses: The number of times to lower and then raise the pipette
                        when picking up a tip, to ensure a good seal (0 [zero]
                        will result in the pipette hovering over the tip but
                        not picking it up--generally not desireable, but could
                        be used for dry-run).
        :type presses: int
        :param increment: The additional distance to travel on each successive
                          press (e.g.: if `presses=3` and `increment=1.0`, then
                          the first press will travel down into the tip by
                          3.5mm, the second by 4.5mm, and the third by 5.5mm).
        :type increment: float

        :returns: This instance
        """
        if location and isinstance(location, types.Location):
            if location.labware.is_labware:
                tiprack = location.labware.as_labware()
                target: Well = tiprack.next_tip(self.channels)  # type: ignore
                if not target:
                    raise OutOfTipsError
            elif location.labware.is_well:
                target = location.labware.as_well()
                tiprack = target.parent
        elif location and isinstance(location, Well):
            tiprack = location.parent
            target = location
        elif not location:
            tiprack, target = next_available_tip(self.starting_tip,
                                                 self.tip_racks,
                                                 self.channels)
        else:
            raise TypeError(
                "If specified, location should be an instance of "
                "types.Location (e.g. the return value from "
                "tiprack.wells()[0].top()) or a Well (e.g. tiprack.wells()[0]."
                " However, it is a {}".format(location))

        assert tiprack.is_tiprack, "{} is not a tiprack".format(str(tiprack))
        validate_tiprack(self.name, tiprack, self._log)
        do_publish(self.broker, cmds.pick_up_tip, self.pick_up_tip,
                   'before', None, None, self, location=target)

        self.move_to(target.top(), publish=False)

        self._implementation.pick_up_tip(
            well=target._impl,
            tip_length=self._tip_length_for(tiprack),
            presses=presses,
            increment=increment
        )
        # Note that the hardware API pick_up_tip action includes homing z after
        do_publish(self.broker, cmds.pick_up_tip, self.pick_up_tip,
                   'after', self, None, self, location=target)

        tiprack.use_tips(target, self.channels)
        self._last_tip_picked_up_from = target

        return self

    @requires_version(2, 0)
    def drop_tip(
            self,
            location: Union[types.Location, Well] = None,
            home_after: bool = True)\
            -> InstrumentContext:
        """
        Drop the current tip.

        If no location is passed, the Pipette will drop the tip into its
        :py:attr:`trash_container`, which if not specified defaults to
        the fixed trash in slot 12.

        The location in which to drop the tip can be manually specified with
        the `location` argument. The `location` argument can be specified in
        several ways:

            - If the only thing to specify is which well into which to drop
              a tip, `location` can be a :py:class:`.Well`. For instance,
              if you have a tip rack in a variable called `tiprack`, you can
              drop a tip into a specific well on that tiprack with the call
              `instr.drop_tip(tiprack.wells()[0])`. This style of call can
              be used to make the robot drop a tip into arbitrary labware.
            - If the position to drop the tip from as well as the
              :py:class:`.Well` to drop the tip into needs to be specified,
              for instance to tell the robot to drop a tip from an unusually
              large height above the tiprack, `location`
              can be a :py:class:`.types.Location`; for instance, you can call
              `instr.drop_tip(tiprack.wells()[0].top())`.

        :param location:
            The location to drop the tip
        :type location:
            :py:class:`.types.Location` or :py:class:`.Well` or None
        :param home_after:
            Whether to home this pipette's plunger after dropping the tip.
            Defaults to ``True``.

            Setting ``home_after=False`` saves waiting a couple of seconds
            after the pipette drops the tip, but risks causing other problems.

            .. warning::
                Only set ``home_after=False`` if:

                * You're using a GEN2 pipette, not a GEN1 pipette.
                * You've tested ``home_after=False`` extensively with your
                  particular pipette and your particular tips.
                * You understand the risks described below.

            The ejector shroud that pops the tip off the end of the pipette is
            driven by the plunger's stepper motor. Sometimes, the strain of
            ejecting the tip can make that motor *skip* and fall out of sync
            with where the robot thinks it is.

            Homing the plunger fixes this, so, to be safe, we normally do it
            after every tip drop.

            If you set ``home_after=False`` to disable homing the plunger, and
            the motor happens to skip, you might see problems like these until
            the next time the plunger is homed:

            * The run might halt with a "hard limit" error message.
            * The pipette might aspirate or dispense the wrong volumes.
            * The pipette might not fully drop subsequent tips.

            GEN1 pipettes are especially vulnerable to this skipping, so you
            should never set ``home_after=False`` with a GEN1 pipette.

            Even on GEN2 pipettes, the motor can still skip. So, always
            extensively test ``home_after=False`` with your particular pipette
            and your particular tips before relying on it.

        :returns: This instance
        """
        if location and isinstance(location, types.Location):
            if location.labware.is_well:
                target = location
            else:
                raise TypeError(
                    "If a location is specified as a types.Location (for "
                    "instance, as the result of a call to "
                    "tiprack.wells()[0].top()) it must be a location "
                    "relative to a well, since that is where a tip is "
                    "dropped. The passed location, however, is in "
                    "reference to {}".format(location.labware))
        elif location and isinstance(location, Well):
            if LabwareLike(location).is_fixed_trash():
                target = location.top()
            else:
                return_height = self._implementation.get_return_height()
                target = determine_drop_target(self.api_version,
                                               location,
                                               return_height)
        elif not location:
            target = self.trash_container.wells()[0].top()
        else:
            raise TypeError(
                "If specified, location should be an instance of "
                "types.Location (e.g. the return value from "
                "tiprack.wells()[0].top()) or a Well (e.g. tiprack.wells()[0]."
                " However, it is a {}".format(location))
        do_publish(self.broker, cmds.drop_tip, self.drop_tip,
                   'before', None, None, self, location=target)
        self.move_to(target, publish=False)

        self._implementation.drop_tip(home_after=home_after)
        do_publish(self.broker, cmds.drop_tip, self.drop_tip,
                   'after', self, None, self, location=target)

        if self.api_version < APIVersion(2, 2) \
                and target.labware.is_well \
                and target.labware.as_well().parent.is_tiprack:
            # If this is a tiprack we can try and add the tip back to the
            # tracker
            try:
                target.labware.as_well().parent.return_tips(
                    target.labware.as_well(), self.channels)
            except AssertionError:
                # Similarly to :py:meth:`return_tips`, the failure case here
                # just means the tip can't be reused, so don't actually stop
                # the protocol
                self._log.exception(f'Could not return tip to {target}')
        self._last_tip_picked_up_from = None
        return self

    @requires_version(2, 0)
    def home(self) -> InstrumentContext:
        """ Home the robot.

        :returns: This instance.
        """
        def home_dummy(mount): pass
        mount_name = self._implementation.get_mount().name.lower()
        do_publish(self.broker, cmds.home, home_dummy,
                   'before', None, None, mount_name)
        self._implementation.home()
        do_publish(self.broker, cmds.home, home_dummy,
                   'after', self, None, mount_name)
        return self

    @requires_version(2, 0)
    def home_plunger(self) -> InstrumentContext:
        """ Home the plunger associated with this mount

        :returns: This instance.
        """
        self._implementation.home_plunger()
        return self

    @publish.both(command=cmds.distribute)
    @requires_version(2, 0)
    def distribute(self,
                   volume: Union[float, Sequence[float]],
                   source: Well,
                   dest: List[Well],
                   *args, **kwargs) -> InstrumentContext:
        """
        Move a volume of liquid from one source to multiple destinations.

        :param volume: The amount of volume to distribute to each destination
                       well.
        :type volume: float or sequence of floats
        :param source: A single well from where liquid will be aspirated.
        :param dest: List of Wells where liquid will be dispensed to.
        :param kwargs: See :py:meth:`transfer`. Some arguments are changed.
                       Specifically, ``mix_after``, if specified, is ignored
                       and ``disposal_volume``, if not specified, is set to the
                       minimum volume of the pipette
        :returns: This instance
        """
        self._log.debug("Distributing {} from {} to {}"
                        .format(volume, source, dest))
        kwargs['mode'] = 'distribute'
        kwargs['disposal_volume'] = kwargs.get(
            'disposal_volume', self.min_volume)
        kwargs['mix_after'] = (0, 0)
        blowout_location = kwargs.get('blowout_location')
        validate_blowout_location(self.api_version, 'distribute', blowout_location)

        return self.transfer(volume, source, dest, **kwargs)

    @publish.both(command=cmds.consolidate)
    @requires_version(2, 0)
    def consolidate(self,
                    volume: Union[float, Sequence[float]],
                    source: List[Well],
                    dest: Well,
                    *args, **kwargs) -> InstrumentContext:
        """
        Move liquid from multiple wells (sources) to a single well(destination)

        :param volume: The amount of volume to consolidate from each source
                       well.
        :type volume: float or sequence of floats
        :param source: List of wells from where liquid will be aspirated.
        :param dest: The single well into which liquid will be dispensed.
        :param kwargs: See :py:meth:`transfer`. Some arguments are changed.
                       Specifically, ``mix_before``, if specified, is ignored
                       and ``disposal_volume`` is ignored and set to 0.
        :returns: This instance
        """
        self._log.debug("Consolidate {} from {} to {}"
                        .format(volume, source, dest))
        kwargs['mode'] = 'consolidate'
        kwargs['mix_before'] = (0, 0)
        kwargs['disposal_volume'] = 0
        blowout_location = kwargs.get('blowout_location')
        validate_blowout_location(self.api_version, 'consolidate', blowout_location)

        return self.transfer(volume, source, dest, **kwargs)

    @publish.both(command=cmds.transfer)  # noqa: C901
    @requires_version(2, 0)
    def transfer(self,
                 volume: Union[float, Sequence[float]],
                 source: AdvancedLiquidHandling,
                 dest: AdvancedLiquidHandling,
                 trash=True,
                 **kwargs) -> InstrumentContext:
        # source: Union[Well, List[Well], List[List[Well]]],
        # dest: Union[Well, List[Well], List[List[Well]]],
        # TODO: Reach consensus on kwargs
        # TODO: Decide if to use a disposal_volume
        # TODO: Accordingly decide if remaining liquid should be blown out to
        # TODO: ..trash or the original well.
        # TODO: What should happen if the user passes a non-first-row well
        # TODO: ..as src/dest *while using multichannel pipette?
        r"""
        Transfer will move a volume of liquid from a source location(s)
        to a dest location(s). It is a higher-level command, incorporating
        other :py:class:`InstrumentContext` commands, like :py:meth:`aspirate`
        and :py:meth:`dispense`, designed to make protocol writing easier at
        the cost of specificity.

        :param volume: The amount of volume to aspirate from each source and
                       dispense to each destination.
                       If volume is a list, each volume will be used for the
                       sources/targets at the matching index. If volumes is a
                       tuple with two elements, like `(20, 100)`, then a list
                       of volumes will be generated with a linear gradient
                       between the two volumes in the tuple.
        :param source: A single well or a list of wells from where liquid
                       will be aspirated.
        :param dest: A single well or a list of wells where liquid
                     will be dispensed to.
        :param \**kwargs: See below

        :Keyword Arguments:

            * *new_tip* (``string``) --

                - 'never': no tips will be picked up or dropped during transfer
                - 'once': (default) a single tip will be used for all commands.
                - 'always': use a new tip for each transfer.

            * *trash* (``boolean``) --
              If `True` (default behavior), tips will be
              dropped in the trash container attached this `Pipette`.
              If `False` tips will be returned to tiprack.

            * *touch_tip* (``boolean``) --
              If `True`, a :py:meth:`touch_tip` will occur following each
              :py:meth:`aspirate` and :py:meth:`dispense`. If set to `False`
              (default behavior), no :py:meth:`touch_tip` will occur.

            * *blow_out* (``boolean``) --
              If `True`, a :py:meth:`blow_out` will occur following each
              :py:meth:`dispense`, but only if the pipette has no liquid left
              in it. If set to `False` (default), no :py:meth:`blow_out` will
              occur.

            * *blowout_location* (``string``) --
                - 'source well': blowout excess liquid into source well
                - 'destintation well': blowout excess liquid into destination
                   well
                - 'trash': blowout excess liquid into the trash
                If no `blowout_location` specified, no `disposal_volume`
                specified, and the pipette contains liquid,
                a :py:meth:`blow_out` will occur into the source well

                If no `blowout_location` specified and either
                `disposal_volume` is specified or the pipette is empty,
                a :py:meth:`blow_out` will occur into the trash

                If `blow_out` is set to `False`, this parameter will be ignored

            * *mix_before* (``tuple``) --
              The tuple, if specified, gives the amount of volume to
              :py:meth:`mix` preceding each :py:meth:`aspirate` during the
              transfer. The tuple is interpreted as (repetitions, volume).

            * *mix_after* (``tuple``) --
              The tuple, if specified, gives the amount of volume to
              :py:meth:`mix` after each :py:meth:`dispense` during the
              transfer. The tuple is interpreted as (repetitions, volume).

            * *disposal_volume* (``float``) --
              (:py:meth:`distribute` only) Volume of liquid to be disposed off
              after distributing. When dispensing multiple times from the same
              tip, it is recommended to aspirate an extra amount of liquid to
              be disposed off after distributing.

            * *carryover* (``boolean``) --
              If `True` (default), any `volume` that exceeds the maximum volume
              of this Pipette will be split into multiple smaller volumes.

            * *gradient* (``lambda``) --
              Function for calculating the curve used for gradient volumes.
              When `volume` is a tuple of length 2, its values are used to
              create a list of gradient volumes. The default curve for this
              gradient is linear (lambda x: x), however a method can be passed
              with the `gradient` keyword argument to create a custom curve.

        :returns: This instance
        """
        self._log.debug("Transfer {} from {} to {}".format(
            volume, source, dest))

        blowout_location = kwargs.get('blowout_location')
        validate_blowout_location(self.api_version, 'transfer', blowout_location)

        kwargs['mode'] = kwargs.get('mode', 'transfer')

        mix_strategy, mix_opts = mix_from_kwargs(kwargs)

        if trash:
            drop_tip = transfers.DropTipStrategy.TRASH
        else:
            drop_tip = transfers.DropTipStrategy.RETURN

        new_tip = kwargs.get('new_tip')
        if isinstance(new_tip, str):
            new_tip = types.TransferTipPolicy[new_tip.upper()]

        blow_out = kwargs.get('blow_out')
        blow_out_strategy = None

        if blow_out and not blowout_location:
            if self.current_volume:
                blow_out_strategy = transfers.BlowOutStrategy.SOURCE
            else:
                blow_out_strategy = transfers.BlowOutStrategy.TRASH
        elif blow_out and blowout_location:
            if blowout_location == 'source well':
                blow_out_strategy = transfers.BlowOutStrategy.SOURCE
            elif blowout_location == 'destination well':
                blow_out_strategy = transfers.BlowOutStrategy.DEST
            elif blowout_location == 'trash':
                blow_out_strategy = transfers.BlowOutStrategy.TRASH

        if new_tip != types.TransferTipPolicy.NEVER:
            tr, next_tip = next_available_tip(self.starting_tip,
                                              self.tip_racks,
                                              self.channels)
            max_volume = min(next_tip.max_volume, self.max_volume)
        else:
            max_volume = self.hw_pipette['working_volume']

        touch_tip = None
        if kwargs.get('touch_tip'):
            touch_tip = transfers.TouchTipStrategy.ALWAYS

        default_args = transfers.Transfer()

        disposal = kwargs.get('disposal_volume')
        if disposal is None:
            disposal = default_args.disposal_volume

        air_gap = kwargs.get('air_gap', default_args.air_gap)
        if air_gap < 0 or air_gap >= max_volume:
            raise ValueError(
                "air_gap must be between 0uL and the pipette's expected "
                f"working volume, {max_volume}uL")

        transfer_args = transfers.Transfer(
            new_tip=new_tip or default_args.new_tip,
            air_gap=air_gap,
            carryover=kwargs.get('carryover') or default_args.carryover,
            gradient_function=(kwargs.get('gradient_function') or
                               default_args.gradient_function),
            disposal_volume=disposal,
            mix_strategy=mix_strategy,
            drop_tip_strategy=drop_tip,
            blow_out_strategy=blow_out_strategy or
            default_args.blow_out_strategy,
            touch_tip_strategy=(touch_tip or
                                default_args.touch_tip_strategy)
        )
        transfer_options = transfers.TransferOptions(transfer=transfer_args,
                                                     mix=mix_opts)
        plan = transfers.TransferPlan(volume, source, dest, self, max_volume,
                                      self.api_version, kwargs['mode'],
                                      transfer_options)
        self._execute_transfer(plan)
        return self

    def _execute_transfer(self, plan: transfers.TransferPlan):
        for cmd in plan:
            getattr(self, cmd['method'])(*cmd['args'], **cmd['kwargs'])

    @requires_version(2, 0)
    def delay(self):
        return self._implementation.delay()

    @requires_version(2, 0)
    def move_to(self,
                location: types.Location,
                force_direct: bool = False,
                minimum_z_height: Optional[float] = None,
                speed: Optional[float] = None,
                publish: bool = True
                ) -> InstrumentContext:
        """ Move the instrument.

        :param location: The location to move to.
        :type location: :py:class:`.types.Location`
        :param force_direct: If set to true, move directly to destination
                        without arc motion.
        :param minimum_z_height: When specified, this Z margin is able to raise
                                 (but never lower) the mid-arc height.
        :param speed: The speed at which to move. By default,
                      :py:attr:`InstrumentContext.default_speed`. This controls
                      the straight linear speed of the motion; to limit
                      individual axis speeds, you can use
                      :py:attr:`.ProtocolContext.max_speeds`.
        :param publish: Whether a call to this function should publish to the
        runlog or not.
        """
        from_loc = self._ctx.location_cache
        if not from_loc:
            from_loc = types.Location(types.Point(0, 0, 0), LabwareLike(None))

        for mod in self._ctx._modules:
            if isinstance(mod, ThermocyclerContext):
                mod.flag_unsafe_move(to_loc=location, from_loc=from_loc)

        if publish:
            do_publish(self.broker, cmds.move_to, self.move_to, 'before',
                       None, None, self, location or self._ctx.location_cache)
        self._implementation.move_to(
            location=location,
            force_direct=force_direct,
            minimum_z_height=minimum_z_height,
            speed=speed
        )
        if publish:
            do_publish(self.broker, cmds.move_to, self.move_to, 'after',
                       None, None, self, location or self._ctx.location_cache)
        return self

    @property  # type: ignore
    @requires_version(2, 0)
    def mount(self) -> str:
        """ Return the name of the mount this pipette is attached to """
        return self._implementation.get_mount().name.lower()

    @property  # type: ignore
    @requires_version(2, 0)
    def speed(self) -> 'PlungerSpeeds':
        """ The speeds (in mm/s) configured for the pipette plunger.

        This is an object with attributes ``aspirate``, ``dispense``, and
        ``blow_out`` holding the plunger speeds for the corresponding
        operation.

        .. note::
            This property is equivalent to :py:attr:`flow_rate`; the only
            difference is the units in which this property is specified.
            Specifying this attribute uses the units of the linear speed of
            the plunger inside the pipette, while :py:attr:`flow_rate` uses
            the units of the volumetric flow rate of liquid into or out of the
            tip. Because :py:attr:`speed` and :py:attr:`flow_rate` modify the
            same values, setting one will override the other.

        For instance, to set the plunger speed during an aspirate action, do

        .. code-block :: python

            instrument.speed.aspirate = 50

        """
        return self._implementation.get_speed()

    @property  # type: ignore
    @requires_version(2, 0)
    def flow_rate(self) -> 'FlowRates':
        """ The speeds (in uL/s) configured for the pipette.

        This is an object with attributes ``aspirate``, ``dispense``, and
        ``blow_out`` holding the flow rates for the corresponding operation.

        .. note::
          This property is equivalent to :py:attr:`speed`; the only
          difference is the units in which this property is specified.
          specifiying this property uses the units of the volumetric flow rate
          of liquid into or out of the tip, while :py:attr:`speed` uses the
          units of the linear speed of the plunger inside the pipette.
          Because :py:attr:`speed` and :py:attr:`flow_rate` modify the
          same values, setting one will override the other.

        For instance, to change the flow rate for aspiration on an instrument
        you would do

        .. code-block :: python

            instrument.flow_rate.aspirate = 50

        """
        return self._implementation.get_flow_rate()

    @property  # type: ignore
    @requires_version(2, 0)
    def type(self) -> str:
        """ One of `'single'` or `'multi'`.
        """
        model = self.name
        if 'single' in model:
            return 'single'
        elif 'multi' in model:
            return 'multi'
        else:
            raise RuntimeError("Bad pipette name: {}".format(model))

    @property  # type: ignore
    @requires_version(2, 0)
    def tip_racks(self) -> List[Labware]:
        """
        The tip racks that have been linked to this pipette.

        This is the property used to determine which tips to pick up next when
        calling :py:meth:`pick_up_tip` without arguments.
        """
        return self._tip_racks

    @tip_racks.setter
    def tip_racks(self, racks: List[Labware]):
        self._tip_racks = racks

    @property  # type: ignore
    @requires_version(2, 0)
    def trash_container(self) -> Labware:
        """ The trash container associated with this pipette.

        This is the property used to determine where to drop tips and blow out
        liquids when calling :py:meth:`drop_tip` or :py:meth:`blow_out` without
        arguments.
        """
        return self._trash

    @trash_container.setter
    def trash_container(self, trash: Labware):
        self._trash = trash

    @property  # type: ignore
    @requires_version(2, 0)
    def name(self) -> str:
        """
        The name string for the pipette (e.g. 'p300_single')
        """
        return self._implementation.get_pipette_name()

    @property  # type: ignore
    @requires_version(2, 0)
    def model(self) -> str:
        """
        The model string for the pipette (e.g. 'p300_single_v1.3')
        """
        return self._implementation.get_model()

    @property  # type: ignore
    @requires_version(2, 0)
    def min_volume(self) -> float:
        return self._implementation.get_min_volume()

    @property  # type: ignore
    @requires_version(2, 0)
    def max_volume(self) -> float:
        """
        The maximum volume, in microliters (µL), that this pipette can hold.

        The maximum volume that you can actually aspirate might be lower than
        this, depending on what kind of tip is attached to this pipette.  For
        example, a P300 Single-Channel pipette always has a ``max_volume`` of
        300 µL, but if it's using a 200 µL filter tip, its usable volume would
        be limited to 200 µL.
        """
        return self._implementation.get_max_volume()

    @property  # type: ignore
    @requires_version(2, 0)
    def current_volume(self) -> float:
        """
        The current amount of liquid, in microliters, held in the pipette.
        """
        return self._implementation.get_current_volume()

    @property  # type: ignore
    @requires_version(2, 7)
    def has_tip(self) -> bool:
        """
        :returns: Whether this instrument has a tip attached or not.
        :type: bool
        """
        return self._implementation.has_tip()

    @property  # type: ignore
    def _has_tip(self) -> bool:
        """
        Internal function used to check whether this instrument has a
        tip attached or not.
        """
        return self._implementation.has_tip()

    @property  # type: ignore
    @requires_version(2, 0)
    def hw_pipette(self) -> PipetteDict:
        """ View the information returned by the hardware API directly.

        :raises: a :py:class:`.types.PipetteNotAttachedError` if the pipette is
                 no longer attached (should not happen).
        """
        return self._implementation.get_pipette()

    @property  # type: ignore
    @requires_version(2, 0)
    def channels(self) -> int:
        """ The number of channels on the pipette. """
        return self._implementation.get_channels()

    @property  # type: ignore
    @requires_version(2, 2)
    def return_height(self) -> float:
        """ The height to return a tip to its tiprack. """
        return self._implementation.get_return_height()

    @property  # type: ignore
    @requires_version(2, 0)
    def well_bottom_clearance(self) -> 'Clearances':
        """ The distance above the bottom of a well to aspirate or dispense.

        This is an object with attributes ``aspirate`` and ``dispense``,
        describing the default heights of the corresponding operation. The
        default is 1.0mm for both aspirate and dispense.

        When :py:meth:`aspirate` or :py:meth:`dispense` is given a
        :py:class:`.Well` rather than a full :py:class:`.Location`, the robot
        will move this distance above the bottom of the well to aspirate or
        dispense.

        To change, set the corresponding attribute. For instance,

        .. code-block:: python

            instr.well_bottom_clearance.aspirate = 1

        """
        return self._implementation.get_well_bottom_clearance()

    def __repr__(self):
        return '<{}: {} in {}>'.format(self.__class__.__name__,
                                       self._implementation.get_model(),
                                       self._implementation.get_mount().name)

    def __str__(self):
        return '{} on {} mount'.format(self.hw_pipette['display_name'],
                                       self.mount)

    @requires_version(2, 7)
    def pair_with(
            self, instrument: InstrumentContext) -> PairedInstrumentContext:
        """ This function allows you to pair both of your pipettes and use
        them simultaneously. The function implicitly decides a primary
        and secondary pipette based on which instrument you call this
        function on.

        :param instrument: The secondary instrument you wish to use

        :raises UnsupportedInstrumentPairingError: If you try to pair
        pipettes that are not currently supported together.
        :returns: PairedInstrumentContext: This is the object you
        will call commands on.

        This function returns a :py:class:`PairedInstrumentContext`.
        The building block commands are the same as an individual pipette's
        building block commands found at :ref:`v2-atomic-commands`,
        and when you want to move pipettes simultaneously you need to use the
        :py:class:`PairedInstrumentContext`.


        Limitations:
        1. This function utilizes a "primary" and "secondary" pipette to make
        positional decisions. The consequence of doing this is that all X & Y
        positions are based on the primary pipette only.
        2. At this time, only pipettes of the same type are supported for
        pipette pairing. This means that you cannot utilize a P1000 Single
        channel and a P300 Single channel at the same time.

        .. code-block :: python
            :substitutions:

            from opentrons import protocol_api

            # metadata
            metadata = {
                'protocolName': 'My Protocol',
                'author': 'Name <email@address.com>',
                'description': 'Simple paired pipette protocol,
                'apiLevel': '|apiLevel|'
            }

            def run(ctx: protocol_api.ProtocolContext):
                right_pipette = ctx.load_instrument(
                    'p300_single_gen2', 'right')
                left_pipette = ctx.load_instrument('p300_single_gen2', 'left')

                # In this scenario, the right pipette is the primary pipette
                # while the left pipette is the secondary pipette. All XY
                # locations will be based on the right pipette.
                right_paired_with_left = right_pipette.pair_with(left_pipette)
                right_paired_with_left.pick_up_tip()
                right_paired_with_left.drop_tip()

                # In this scenario, the left pipette is the primary pipette
                # while the right pipette is the secondary pipette. All XY
                # locations will be based on the left pipette.
                left_paired_with_right = left_pipette.pair_with(right_pipette)
                left_paired_with_right.pick_up_tip()
                left_paired_with_right.drop_tip()

        .. note::

            Before using this method, you should seriously consider whether
            this is the best fit for your use-case especially given the
            limitations listed above.
        """
        if instrument.name != self.name:
            raise UnsupportedInstrumentPairingError(
                'At this time, you cannot pair'
                f'{instrument.name} with {self.name}')

        return PairedInstrumentContext(
            primary_instrument=self,
            secondary_instrument=instrument,
            ctx=self._ctx,
            pair_policy=PipettePair.of_mount(self._implementation.get_mount()),
            api_version=self.api_version,
            hardware_manager=self._ctx._implementation.get_hardware(),
            trash=self.trash_container,
            log_parent=self._log
        )

    def _tip_length_for(self, tiprack: Labware) -> float:
        """ Get the tip length, including overlap, for a tip from this rack """
        return tip_length_for(self.hw_pipette, tiprack)



from typing import Dict, Optional, Any, List
from dataclasses import dataclass, asdict, field
import json
from pathlib import Path

from opentrons.config import robot_configs
from opentrons.config.types import RobotConfig
from opentrons.types import Mount
from opentrons.hardware_control import API


# Name and kwargs for a module function
@dataclass(frozen=True)
class ModuleCall:
    function_name: str
    args: List[Any] = field(default_factory=list)
    kwargs: Dict[str, Any] = field(default_factory=dict)


@dataclass(frozen=True)
class SimulatorSetup:
    attached_instruments: Dict[Mount, Dict[str, Optional[str]]] =\
        field(default_factory=dict)
    attached_modules: Dict[str, List[ModuleCall]] = field(default_factory=dict)
    config: Optional[RobotConfig] = None
    strict_attached_instruments: bool = True


async def create_simulator(setup: SimulatorSetup, loop=None) -> API:
    """Create a simulator"""
    simulator = await API.build_hardware_simulator(
        attached_instruments=setup.attached_instruments,
        attached_modules=list(setup.attached_modules.keys()),
        config=setup.config,
        strict_attached_instruments=setup.strict_attached_instruments,
        loop=loop,
    )

    for attached_module in simulator.attached_modules:
        calls = setup.attached_modules[attached_module.name()]
        for call in calls:
            f = getattr(attached_module, call.function_name)
            await f(*call.args, **call.kwargs)

    return simulator


async def load_simulator(path: Path, loop=None) -> API:
    """Create a simulator from a JSON file."""
    return await create_simulator(setup=load_simulator_setup(path),
                                  loop=loop)


def save_simulator_setup(simulator_setup: SimulatorSetup, path: Path):
    """Write a simulator setup to a file."""
    as_dict = asdict(simulator_setup)
    as_dict = {k: _prepare_for_dict(k, v) for (k, v) in as_dict.items()}
    with path.open('w') as f:
        json.dump(as_dict, f)


def load_simulator_setup(path: Path) -> SimulatorSetup:
    """Load a simulator setup from a file."""
    with path.open() as f:
        obj = json.load(f)
        return SimulatorSetup(
            **{k: _prepare_for_simulator_setup(k, v) for (k, v)
               in obj.items()})


def _prepare_for_dict(key, value):
    """Convert an element in SimulatorSetup to be a serializable dict"""
    if key == 'attached_instruments' and value:
        return {mount.name.lower(): data for (mount, data) in value.items()}
    return value


def _prepare_for_simulator_setup(key, value):
    """Convert value to a SimulatorSetup"""
    if key == 'attached_instruments' and value:
        return {Mount[mount.upper()]: data for (mount, data) in value.items()}
    if key == 'config' and value:
        return robot_configs.build_config(value)
    if key == 'attached_modules' and value:
        return {k: [ModuleCall(**data) for data in v] for (k, v)
                in value.items()}
    return value



import asyncio
import logging
from typing import Callable, Dict, Tuple

from opentrons.hardware_control.types import BoardRevision, DoorState
from .types import gpio_group, GPIOPin

MODULE_LOG = logging.getLogger(__name__)


class SimulatingGPIOCharDev:
    def __init__(self, chip_name: str):
        self._chip = chip_name
        self._board_rev = BoardRevision.UNKNOWN
        self._lines = self._initialize()

    @property
    def chip(self) -> str:
        return self._chip

    @property
    def lines(self) -> Dict[str, int]:
        return self._lines

    @property
    def board_rev(self) -> BoardRevision:
        return self._board_rev

    @board_rev.setter
    def board_rev(self, boardrev: BoardRevision):
        self._board_rev = boardrev

    def _initialize(self) -> Dict[str, int]:
        lines = {}
        for pin in gpio_group.pins:
            lines[pin.name] = pin.by_board_rev(self.board_rev)
        self._initialize_values(list(lines.keys()))
        return lines

    def _initialize_values(self, names):
        self._values: Dict[str, int] = {}
        for name in names:
            self._values[name] = 1

    async def setup(self):
        pass

    def config_by_board_rev(self):
        self.board_rev = BoardRevision.by_bits(
            self.read_revision_bits())

    def set_high(self, output_pin: GPIOPin):
        self._values[output_pin.name] = 1

    def set_low(self, output_pin: GPIOPin):
        self._values[output_pin.name] = 0

    def set_button_light(self,
                         red: bool = False,
                         green: bool = False,
                         blue: bool = False):
        pass

    def set_rail_lights(self, on: bool = True):
        pass

    def set_reset_pin(self, on: bool = True):
        pass

    def set_isp_pin(self, on: bool = True):
        pass

    def set_halt_pin(self, on: bool = True):
        pass

    def _read(self, input_pin: GPIOPin) -> int:
        return self._values[input_pin.name]

    def get_button_light(self) -> Tuple[bool, bool, bool]:
        return (bool(self._read(gpio_group.red_button)),
                bool(self._read(gpio_group.green_button)),
                bool(self._read(gpio_group.blue_button)))

    def get_rail_lights(self) -> bool:
        return bool(self._read(gpio_group.frame_leds))

    def read_button(self) -> bool:
        # button is normal-HIGH, so invert
        return not bool(self._read(gpio_group.button_input))

    def read_window_switches(self) -> bool:
        return bool(self._read(gpio_group.window_door_sw))

    def read_top_window_switch(self) -> bool:
        return bool(self._read(gpio_group.window_sw_filt))

    def read_front_door_switch(self) -> bool:
        return bool(self._read(gpio_group.door_sw_filt))

    def read_revision_bits(self) -> Tuple[bool, bool]:
        return (bool(self._read(gpio_group.rev_0)),
                bool(self._read(gpio_group.rev_1)))

    def get_door_state(self) -> DoorState:
        val = self.read_window_switches()
        if val:
            return DoorState.CLOSED
        else:
            return DoorState.OPEN

    def start_door_switch_watcher(
            self, loop: asyncio.AbstractEventLoop,
            update_door_state: Callable[[DoorState], None]):
        current_door_value = self.read_window_switches()
        if current_door_value == 0:
            update_door_state(DoorState.OPEN)
        else:
            update_door_state(DoorState.CLOSED)

    def release_line(self, pin: GPIOPin):
        self.lines.pop(pin.name)

    def stop_door_switch_watcher(self, loop: asyncio.AbstractEventLoop):
        pass



from .helpers import stringify_location
from . import types as command_types

from typing import Union, Sequence, Optional, TYPE_CHECKING

from opentrons.protocols.api_support.util import FlowRates

if TYPE_CHECKING:
    from opentrons.protocol_api.instrument_context import InstrumentContext
    from opentrons.protocol_api.labware import Well
    from opentrons.types import Location


Apiv2Locations = Sequence[Union['Location', 'Well']]
Apiv2Instruments = Sequence['InstrumentContext']


def combine_locations(location: Sequence) -> str:
    if len(location) > 1:
        loc1 = stringify_location(location[0])
        loc2 = stringify_location(location[1])
        return f'{loc1} and {loc2}'
    elif len(location) == 1:
        loc1 = stringify_location(location[0])
        return f'{loc1}'
    else:
        return ''


def paired_aspirate(
        instruments: Apiv2Instruments, volume: float,
        locations: Apiv2Locations, rate: float,
        pub_type: str) -> command_types.AspirateCommand:
    loc_text = combine_locations(locations)
    flow_rate = min(
        rate * FlowRates(instr._implementation).aspirate
        for instr in instruments)
    text_type = f'{pub_type}: Aspirating '
    text_content = f'{volume} uL from {loc_text} at {flow_rate} uL/sec'
    text = text_type + text_content
    return {
        'name': command_types.ASPIRATE,
        'payload': {
            'instruments': instruments,
            'volume': volume,
            'locations': locations,
            'rate': rate,
            'text': text
        }
    }


def paired_dispense(
        instruments: Apiv2Instruments, volume: float,
        locations: Apiv2Locations, rate: float,
        pub_type: str) -> command_types.DispenseCommand:
    loc_text = combine_locations(locations)
    flow_rate = min(
        rate * FlowRates(instr._implementation).dispense
        for instr in instruments)
    text_type = f'{pub_type}: Dispensing '
    text_content = f'{volume} uL into {loc_text} at {flow_rate} uL/sec'
    text = text_type + text_content
    return {
        'name': command_types.DISPENSE,
        'payload': {
            'instruments': instruments,
            'volume': volume,
            'locations': locations,
            'rate': rate,
            'text': text
        }
    }


def paired_mix(
        instruments: Apiv2Instruments, locations: Optional[Apiv2Locations],
        repetitions: int, volume: float, pub_type: str) -> command_types.MixCommand:
    text_type = f'{pub_type}: Mixing '
    text_content = '{repetitions} times with a volume of {volume} ul'
    text = text_type + text_content
    return {
        'name': command_types.MIX,
        'payload': {
            'instruments': instruments,
            'locations': locations,
            'volume': volume,
            'repetitions': repetitions,
            'text': text
        }
    }


def paired_blow_out(
        instruments: Apiv2Instruments,
        locations: Optional[Apiv2Locations],
        pub_type: str) -> command_types.BlowOutCommand:
    text = f'{pub_type}: Blowing out'

    if locations is not None:
        location_text = combine_locations(locations)
        text += f' at {location_text}'

    return {
        'name': command_types.BLOW_OUT,
        'payload': {
            'instruments': instruments,
            'locations': locations,
            'text': text
        }
    }


def paired_touch_tip(
        instruments: Apiv2Instruments,
        locations: Optional[Apiv2Locations],
        pub_type: str) -> command_types.TouchTipCommand:
    text = f'{pub_type}: Touching tip'

    if locations is not None:
        location_text = combine_locations(locations)
        text += f' at {location_text}'
    return {
        'name': command_types.TOUCH_TIP,
        'payload': {
            'instruments': instruments,
            'locations': locations,
            'text': text
        }
    }


def air_gap() -> command_types.AirGapCommand:
    text = 'Air gap'
    return {
        'name': command_types.AIR_GAP,
        'payload': {
            'text': text
        }
    }


def return_tip() -> command_types.ReturnTipCommand:
    text = 'Returning tip'
    return {
        'name': command_types.RETURN_TIP,
        'payload': {
            'text': text
        }
    }


def paired_pick_up_tip(
        instruments: Apiv2Instruments,
        locations: Apiv2Locations, pub_type: str) -> command_types.PickUpTipCommand:
    location_text = combine_locations(locations)
    text = f'{pub_type}: Picking up tip from {location_text}'
    return {
        'name': command_types.PICK_UP_TIP,
        'payload': {
            'instruments': instruments,
            'locations': locations,
            'text': text
        }
    }


def paired_drop_tip(
        instruments: Apiv2Instruments,
        locations: Apiv2Locations, pub_type: str) -> command_types.DropTipCommand:
    location_text = combine_locations(locations)
    text = f'{pub_type}: Dropping tip into {location_text}'
    return {
        'name': command_types.DROP_TIP,
        'payload': {
            'instruments': instruments,
            'locations': locations,
            'text': text
        }
    }


def paired_move_to(
        instruments: Apiv2Instruments,
        locations: Apiv2Locations, pub_type: str) -> command_types.MoveToCommand:
    location_text = combine_locations(locations)
    text = f'{pub_type}: Moving to {location_text}'
    return {
        'name': command_types.MOVE_TO,
        'payload': {
            'instruments': instruments,
            'locations': locations,
            'text': text
        }
    }



"""Geometry state store and getters."""
from dataclasses import dataclass
from typing import Optional
from typing_extensions import final

from opentrons_shared_data.deck.dev_types import DeckDefinitionV2, SlotDefV2
from opentrons.types import Point, DeckSlotName
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.protocols.geometry.deck import FIXED_TRASH_ID

from .. import errors
from ..types import WellLocation, WellOrigin
from .substore import Substore, CommandReactive
from .labware import LabwareStore, LabwareData


DEFAULT_TIP_DROP_HEIGHT_FACTOR = 0.5

# TODO(mc, 2020-11-12): reconcile this data structure with WellGeometry


@final
@dataclass(frozen=True)
class TipGeometry:
    """Tip geometry data."""

    effective_length: float
    diameter: float
    volume: int


class GeometryState:
    """Geometry getters."""

    _deck_definition: DeckDefinitionV2
    _labware_store: LabwareStore

    def __init__(
        self,
        deck_definition: DeckDefinitionV2,
        labware_store: LabwareStore
    ) -> None:
        """Initialize a GeometryState instance."""
        self._deck_definition = deck_definition
        self._labware_store = labware_store

    def get_deck_definition(self) -> DeckDefinitionV2:
        """Get the current deck definition."""
        return self._deck_definition

    def get_slot_definition(self, slot: DeckSlotName) -> SlotDefV2:
        """Get the current deck definition."""
        deck_def = self.get_deck_definition()

        for slot_def in deck_def["locations"]["orderedSlots"]:
            if slot_def["id"] == str(slot):
                return slot_def

        raise errors.SlotDoesNotExistError(
            f"Slot ID {slot} does not exist in deck {deck_def['otId']}"
        )

    def get_slot_position(self, slot: DeckSlotName) -> Point:
        """Get the position of a deck slot."""
        slot_def = self.get_slot_definition(slot)
        position = slot_def["position"]

        return Point(x=position[0], y=position[1], z=position[2])

    def get_labware_highest_z(self, labware_id: str) -> float:
        """Get the highest Z-point of a labware."""
        labware_data = self._labware_store.state.get_labware_data_by_id(
            labware_id
        )

        return self._get_highest_z_from_labware_data(labware_data)

    def get_all_labware_highest_z(self) -> float:
        """Get the highest Z-point of a labware."""
        return max([
            self._get_highest_z_from_labware_data(lw_data)
            for uid, lw_data in self._labware_store.state.get_all_labware()
        ])

    def get_labware_parent_position(
        self,
        labware_id: str
    ) -> Point:
        """Get the position of the labware's parent slot (deck or module)."""
        labware_data = self._labware_store.state.get_labware_data_by_id(labware_id)
        slot_pos = self.get_slot_position(labware_data.location.slot)

        return slot_pos

    def get_labware_origin_position(
        self,
        labware_id: str
    ) -> Point:
        """Get the position of the labware's origin, without calibration."""
        labware_data = self._labware_store.state.get_labware_data_by_id(labware_id)
        slot_pos = self.get_slot_position(labware_data.location.slot)
        origin_offset = self._labware_store.state.get_definition_by_uri(
            labware_data.uri
        ).cornerOffsetFromSlot

        return Point(
            x=slot_pos.x + origin_offset.x,
            y=slot_pos.y + origin_offset.y,
            z=slot_pos.z + origin_offset.z
        )

    def get_labware_position(
        self,
        labware_id: str
    ) -> Point:
        """Get the calibrated origin of the labware."""
        labware_data = self._labware_store.state.get_labware_data_by_id(labware_id)
        origin_pos = self.get_labware_origin_position(labware_id=labware_id)
        cal_offset = labware_data.calibration

        return Point(
            x=origin_pos.x + cal_offset[0],
            y=origin_pos.y + cal_offset[1],
            z=origin_pos.z + cal_offset[2]
        )

    def get_well_position(
        self,
        labware_id: str,
        well_name: str,
        well_location: Optional[WellLocation] = None,
    ) -> Point:
        """Get the absolute position of a well in a labware."""
        labware_pos = self.get_labware_position(
            labware_id
        )
        well_def = self._labware_store.state.get_well_definition(
            labware_id,
            well_name
        )
        well_depth = well_def.depth

        if well_location is not None:
            offset = well_location.offset

            if well_location.origin == WellOrigin.TOP:
                offset = (offset[0], offset[1], offset[2] + well_depth)

        else:
            offset = (0, 0, well_depth)

        return Point(
            x=labware_pos.x + offset[0] + well_def.x,
            y=labware_pos.y + offset[1] + well_def.y,
            z=labware_pos.z + offset[2] + well_def.z,
        )

    def _get_highest_z_from_labware_data(self, lw_data: LabwareData) -> float:
        z_dim = self._labware_store.state.get_definition_by_uri(
            lw_data.uri
        ).dimensions.zDimension
        slot_pos = self.get_slot_position(lw_data.location.slot)

        return z_dim + slot_pos[2] + lw_data.calibration[2]

    # TODO(mc, 2020-11-12): reconcile with existing protocol logic and include
    # data from tip-length calibration once v4.0.0 is in `edge`
    def get_effective_tip_length(
        self,
        labware_id: str,
        pipette_config: PipetteDict
    ) -> float:
        """Given a labware and a pipette's config, get the effective tip length.

        Effective tip length is the nominal tip length less the distance the
        tip overlaps with the pipette nozzle.
        """
        labware_uri = self._labware_store.state.get_definition_uri(labware_id)
        nominal_length = self._labware_store.state.get_tip_length(labware_id)
        overlap_config = pipette_config["tip_overlap"]
        default_overlap = overlap_config.get("default", 0)
        overlap = overlap_config.get(labware_uri, default_overlap)

        return nominal_length - overlap

    # TODO(mc, 2020-11-12): reconcile with existing geometry logic
    def get_tip_geometry(
        self,
        labware_id: str,
        well_name: str,
        pipette_config: PipetteDict
    ) -> TipGeometry:
        """Given a labware, well, and hardware pipette config, get the tip geometry.

        Tip geometry includes effective tip length, tip diameter, and tip volume,
        which is all data required by the hardware controller for proper tip handling.
        """
        effective_length = self.get_effective_tip_length(labware_id, pipette_config)
        well_def = self._labware_store.state.get_well_definition(labware_id, well_name)

        if well_def.shape != "circular":
            raise errors.LabwareIsNotTipRackError(
                f"Well {well_name} in labware {labware_id} is not circular."
            )

        return TipGeometry(
            effective_length=effective_length,
            diameter=well_def.diameter,  # type: ignore[arg-type]
            # TODO(mc, 2020-11-12): WellDefinition type says totalLiquidVolume
            #  is a float, but hardware controller expects an int
            volume=int(well_def.totalLiquidVolume),
        )

    # TODO(mc, 2020-11-12): support pre-PAPIv2.2/2.3 behavior of dropping the tip
    # 10mm above well bottom
    def get_tip_drop_location(
        self,
        labware_id: str,
        pipette_config: PipetteDict
    ) -> WellLocation:
        """Get tip drop location given labware and hardware pipette."""
        # return to top if labware is fixed trash
        if labware_id == FIXED_TRASH_ID:
            return WellLocation()

        nominal_length = self._labware_store.state.get_tip_length(labware_id)
        offset_factor = pipette_config["return_tip_height"]

        return WellLocation(offset=(0, 0, -nominal_length * offset_factor))


class GeometryStore(Substore[GeometryState], CommandReactive):
    """Geometry state container."""

    _state: GeometryState

    def __init__(
        self,
        deck_definition: DeckDefinitionV2,
        labware_store: LabwareStore
    ) -> None:
        """Initialize a geometry store and its state."""
        self._state = GeometryState(
            deck_definition=deck_definition,
            labware_store=labware_store,
        )



"""The interface that implements InstrumentContext."""

from abc import abstractmethod
from typing import List, Dict

from opentrons.protocols.geometry.deck_item import DeckItem
from opentrons.protocols.geometry.labware_geometry import AbstractLabwareGeometry
from opentrons.protocols.api_support.tip_tracker import TipTracker
from opentrons.protocols.context.well import WellImplementation
from opentrons.protocols.api_support.well_grid import WellGrid
from opentrons.types import Point
from opentrons_shared_data.labware.dev_types import (
    LabwareParameters, LabwareDefinition)


class AbstractLabware(DeckItem):
    """Abstract base class of Labware Implementations"""

    @abstractmethod
    def get_uri(self) -> str:
        ...

    @abstractmethod
    def get_display_name(self) -> str:
        ...

    @abstractmethod
    def get_name(self) -> str:
        ...

    @abstractmethod
    def set_name(self, new_name: str) -> None:
        ...

    @abstractmethod
    def get_definition(self) -> LabwareDefinition:
        ...

    @abstractmethod
    def get_parameters(self) -> LabwareParameters:
        ...

    @abstractmethod
    def get_quirks(self) -> List[str]:
        ...

    @abstractmethod
    def set_calibration(self, delta: Point) -> None:
        ...

    @abstractmethod
    def get_calibrated_offset(self) -> Point:
        ...

    @abstractmethod
    def is_tiprack(self) -> bool:
        ...

    @abstractmethod
    def get_tip_length(self) -> float:
        ...

    @abstractmethod
    def set_tip_length(self, length: float):
        ...

    @abstractmethod
    def reset_tips(self) -> None:
        ...

    @abstractmethod
    def get_tip_tracker(self) -> TipTracker:
        ...

    @abstractmethod
    def get_well_grid(self) -> WellGrid:
        ...

    @abstractmethod
    def get_wells(self) -> List[WellImplementation]:
        ...

    @abstractmethod
    def get_wells_by_name(self) -> Dict[str, WellImplementation]:
        ...

    @abstractmethod
    def get_geometry(self) -> AbstractLabwareGeometry:
        ...



"""Protocol engine module.

The protocol_engine module contains the logic necessary to take a stream of
protocol commands, issued by some arbitrary protocol runner, and turn it into
protocol state and side-effects like robot movements.
"""

from .protocol_engine import ProtocolEngine
from .state import StateStore, StateView, LabwareData
from .execution import CommandHandlers
from .resources import ResourceProviders
from .types import (
    DeckLocation,
    DeckSlotLocation,
    Dimensions,
    PipetteName,
    WellLocation,
    WellOrigin,
)

__all__ = [
    # main export
    "ProtocolEngine",
    # state interfaces and models
    "StateStore",
    "StateView",
    "LabwareData",
    # command execution interfaces
    "CommandHandlers",
    # resource management interfaces
    "ResourceProviders",
    # type definitions and other value models
    "DeckLocation",
    "DeckSlotLocation",
    "Dimensions",
    "PipetteName",
    "WellLocation",
    "WellOrigin",
]



"""File runner interfaces for Python protocols."""
from .abstract_file_runner import AbstractFileRunner


class PythonFileRunner(AbstractFileRunner):
    """Python protocol file runner."""

    def load(self) -> None:
        """Prepare to run the Python protocol file."""
        raise NotImplementedError("Python protocol loading not implemented")

    def play(self) -> None:
        """Start (or un-pause) running the Python protocol file."""
        raise NotImplementedError()

    def pause(self) -> None:
        """Pause the running Python protocol file's execution."""
        raise NotImplementedError()

    def stop(self) -> None:
        """Cancel the running Python protocol file."""
        raise NotImplementedError()



"""Pick up tip command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel

from .command import CommandImplementation, CommandHandlers
from .pipetting_common import BasePipettingRequest


class PickUpTipRequest(BasePipettingRequest):
    """A request to move a pipette to a specific well."""

    def get_implementation(self) -> PickUpTipImplementation:
        """Get the pick up tip request's command implementation."""
        return PickUpTipImplementation(self)


class PickUpTipResult(BaseModel):
    """Result data from the execution of a PickUpTipRequest."""

    pass


class PickUpTipImplementation(
    CommandImplementation[PickUpTipRequest, PickUpTipResult]
):
    """Pick up tip command implementation."""

    async def execute(self, handlers: CommandHandlers) -> PickUpTipResult:
        """Move to and pick up a tip using the requested pipette."""
        await handlers.pipetting.pick_up_tip(
            pipette_id=self._request.pipetteId,
            labware_id=self._request.labwareId,
            well_name=self._request.wellName,
        )

        return PickUpTipResult()



import logging
import os
import shutil
from enum import Enum
from pathlib import Path
from typing import NamedTuple, Dict, Set

from opentrons.config import IS_ROBOT
from opentrons.calibration_storage import delete

DATA_BOOT_D = Path('/data/boot.d')

log = logging.getLogger(__name__)


class UnrecognizedOption(Exception):
    pass


class CommonResetOption(NamedTuple):
    name: str
    description: str


class ResetOptionId(str, Enum):
    """The available reset options"""
    labware_calibration = 'labwareCalibration'
    boot_scripts = 'bootScripts'
    deck_calibration = 'deckCalibration'
    pipette_offset = 'pipetteOffsetCalibrations'
    tip_length_calibrations = 'tipLengthCalibrations'


_settings_reset_options = {
    ResetOptionId.labware_calibration:
        CommonResetOption(
            name='Labware Calibration',
            description='Clear labware calibration'
        ),
    ResetOptionId.boot_scripts:
        CommonResetOption(
            name='Boot Scripts',
            description='Clear custom boot scripts'
        ),
    ResetOptionId.deck_calibration:
        CommonResetOption(
            name='Deck Calibration',
            description='Clear deck calibration (will also clear pipette '
                        'offset)'
        ),
    ResetOptionId.pipette_offset:
        CommonResetOption(
            name='Pipette Offset Calibrations',
            description='Clear pipette offset calibrations'
        ),
    ResetOptionId.tip_length_calibrations:
        CommonResetOption(
            name='Tip Length Calibrations',
            description='Clear tip length calibrations (will also clear '
                        'pipette offset)'
        )
}


def reset_options() -> Dict[ResetOptionId, CommonResetOption]:
    return _settings_reset_options


def reset(options: Set[ResetOptionId]) -> None:
    """
    Execute a reset of the requested parts of the user configuration.

    :param options: the parts to reset
    """
    log.info("Reset requested for %s", options)

    if ResetOptionId.labware_calibration in options:
        reset_labware_calibration()

    if ResetOptionId.boot_scripts in options:
        reset_boot_scripts()

    if ResetOptionId.deck_calibration in options:
        reset_deck_calibration()

    if ResetOptionId.pipette_offset in options:
        reset_pipette_offset()

    if ResetOptionId.tip_length_calibrations in options:
        reset_tip_length_calibrations()


def reset_boot_scripts():
    if IS_ROBOT:
        if os.path.exists(DATA_BOOT_D):
            shutil.rmtree(DATA_BOOT_D)
    else:
        log.debug(f'Not on pi, not removing {DATA_BOOT_D}')


def reset_deck_calibration():
    delete.delete_robot_deck_attitude()
    delete.clear_pipette_offset_calibrations()


def reset_pipette_offset():
    delete.clear_pipette_offset_calibrations()


def reset_tip_length_calibrations():
    delete.clear_tip_length_calibration()
    delete.clear_pipette_offset_calibrations()


def reset_labware_calibration():
    delete.clear_calibrations()



import hashlib
import logging
import os
import shutil
from typing import Generator, Optional, Dict, Any
from dataclasses import dataclass

from opentrons.config import CONFIG
from opentrons.system import nmcli

log = logging.getLogger(__name__)


class ConfigureArgsError(Exception):
    pass


EAP_CONFIG_SHAPE = {
    'options': [
        {'name': method.qualified_name(),
         'displayName': method.display_name(),
         'options': [{k: v for k, v in arg.items()
                      if k in ['name',
                               'displayName',
                               'required',
                               'type']}
                     for arg in method.args()]}
        for method in nmcli.EAP_TYPES]
}


@dataclass(frozen=True)
class Key:
    directory: str
    file: str


@dataclass(frozen=True)
class AddKeyResult:
    created: bool
    key: Key


def add_key(key_file_name: str, key_contents: bytes) -> AddKeyResult:
    """
    Add a key file (for later use in EAP config) to the system.
    """
    keys_dir = CONFIG['wifi_keys_dir']
    hasher = hashlib.sha256()
    hasher.update(key_contents)
    key_hash = hasher.hexdigest()
    if key_hash in os.listdir(keys_dir):
        files = os.listdir(os.path.join(keys_dir, key_hash))
        if files:
            return AddKeyResult(created=False,
                                key=Key(directory=key_hash,
                                        file=files[0]))
        else:
            log.warning(
                "Key directory with nothing in it: {}"
                .format(key_hash))
            os.rmdir(os.path.join(keys_dir, key_hash))

    key_hash_path = os.path.join(keys_dir, key_hash)
    os.mkdir(key_hash_path)
    with open(os.path.join(key_hash_path,
                           os.path.basename(key_file_name)), 'wb') as f:
        f.write(key_contents)
    return AddKeyResult(created=True,
                        key=Key(directory=key_hash,
                                file=key_file_name))


def list_keys() -> Generator[Key, None, None]:
    """
    List wifi keys known to the system.

    :return: A generator yielding Key objects
    """
    keys_dir = CONFIG['wifi_keys_dir']
    # TODO(mc, 2018-10-24): add last modified info to keys for sort purposes
    for path in os.listdir(keys_dir):
        full_path = os.path.join(keys_dir, path)
        if os.path.isdir(full_path):
            in_path = os.listdir(full_path)
            if len(in_path) > 1:
                log.warning("Garbage in key dir for key {}".format(path))
            yield Key(directory=path,
                      file=in_path[0])
        else:
            log.warning("Garbage in wifi keys dir: {}".format(full_path))


def remove_key(requested_hash: str) -> Optional[str]:
    """
    Try to delete key file

    :param requested_hash: The hash to delete
    :return: The name of the deleted file or None if not found
    """
    keys_dir = CONFIG['wifi_keys_dir']
    available_keys = os.listdir(keys_dir)
    if requested_hash not in available_keys:
        return None
    key_path = os.path.join(keys_dir, requested_hash)
    name = os.listdir(key_path)[0]
    shutil.rmtree(key_path)
    return name


def get_key_file(key: str) -> str:
    """
    Get the full path of a key file

    :param key: The key to look for
    :return: the path
    """
    keys_dir = CONFIG['wifi_keys_dir']
    available_keys = os.listdir(keys_dir)
    if key not in available_keys:
        raise ConfigureArgsError(f'Key ID {key} is not valid on the system')
    files_in_dir = os.listdir(os.path.join(keys_dir, key))
    if len(files_in_dir) > 1:
        raise OSError(
            f'Key ID {key} has multiple files, try deleting and re-uploading'
        )
    return os.path.join(keys_dir, key, files_in_dir[0])


def _eap_check_no_extra_args(
        config: Dict[str, Any], options: Any):
    # options is an Any because the type annotation for EAP_CONFIG_SHAPE itself
    # can’t quite express the type properly because of the inference from the
    # dict annotation.
    """Check for args that are not required for this method (to aid debugging)
    ``config`` should be the user config.
    ``options`` should be the options sub-member for the eap method.

    Before this method is called, the validity of the 'eapType' key should be
    established.
    """
    arg_names = [k for k in config.keys() if k != 'eapType']
    valid_names = [o['name'] for o in options]
    for an in arg_names:
        if an not in valid_names:
            raise ConfigureArgsError(
                'Option {} is not valid for EAP method {}'
                .format(an, config['eapType']))


def _eap_check_option_ok(opt: Dict[str, str], config: Dict[str, Any]):
    """
    Check that a given EAP option is in the user config (if required)
    and, if specified, is the right type.

    ``opt`` should be an options dict from EAP_CONFIG_SHAPE.
    ``config`` should be the user config dict.

    Before this method is called, the validity of the eapType key should be
    established.
    """
    if opt['name'] not in config:
        if opt['required']:
            raise ConfigureArgsError(
                'Required argument {} for eap method {} not present'
                .format(opt['displayName'], config['eapType']))
        else:
            return
    name = opt['name']
    o_type = opt['type']
    arg = config[name]
    if name in config:
        if o_type in ('string', 'password') and not isinstance(arg, str):
            raise ConfigureArgsError('Option {} should be a str'
                                     .format(name))
        elif o_type == 'file' and not isinstance(arg, str):
            raise ConfigureArgsError('Option {} must be a str'
                                     .format(name))


def eap_check_config(eap_config: Dict[str, Any]) -> Dict[str, Any]:
    """Check the eap specific args, and replace values where needed."""
    eap_type = eap_config.get('eapType')
    for method in EAP_CONFIG_SHAPE['options']:
        if method['name'] == eap_type:
            options = method['options']
            break
    else:
        raise ConfigureArgsError('EAP method {} is not valid'.format(eap_type))

    _eap_check_no_extra_args(eap_config, options)

    for opt in options:  # type: ignore
        # Ignoring most types to do with EAP_CONFIG_SHAPE because of issues
        # wth type inference for dict comprehensions
        _eap_check_option_ok(opt, eap_config)
        if opt['type'] == 'file' and opt['name'] in eap_config:
            # Special work for file: rewrite from key id to path
            eap_config[opt['name']] = get_key_file(eap_config[opt['name']])
    return eap_config



# noqa: D100

from __future__ import annotations

from typing import Optional, Union, Sequence, List

from .labware import Well, Labware

from opentrons import APIVersion, types
from opentrons.hardware_control.dev_types import PipetteDict

# todo(mm, 2021-04-09): Duplicate these classes in this package to
# decouple from the v2 opentrons.protocol_api?
from opentrons.protocol_api import PairedInstrumentContext
from opentrons.protocol_api.instrument_context import AdvancedLiquidHandling
from opentrons.protocol_engine import WellLocation, WellOrigin
from opentrons.protocol_engine.clients import SyncClient as ProtocolEngineClient

# todo(mm, 2021-04-09): How customer-facing are these classes? Should they be
# accessible and documented as part of this package?
from opentrons.protocols.api_support.util import PlungerSpeeds, FlowRates, Clearances


# todo(mm, 2021-04-09): Can/should we remove the word "Context" from the name?
class PipetteContext:  # noqa: D101
    def __init__(
        self,
        engine_client: ProtocolEngineClient,
        pipette_id: str,
    ) -> None:
        """Initialize a PipetteContext API provider.

        You should not need to call this constructor yourself. The system will
        create a PipetteContext for you when you call :py:meth:`load_pipette`.

        Args:
            engine_client: A client to access protocol state and execute commands.
            pipette_id: The pipette's identifier in commands and protocol state.
        """
        self._engine_client = engine_client
        self._pipette_id = pipette_id

    def __hash__(self) -> int:
        """Get hash.

        Uses the pipette instance's unique identifier in protocol state.
        """
        return hash(self._pipette_id)

    def __eq__(self, other: object) -> bool:
        """Compare for object equality.

        Checks that other object is a `PipetteContext` and has the same identifier.
        """
        return (
            isinstance(other, PipetteContext) and self._pipette_id == other._pipette_id
        )

    def __repr__(self) -> str:  # noqa: D105
        raise NotImplementedError()

    def __str__(self) -> str:  # noqa: D105
        raise NotImplementedError()

    @property
    def api_version(self) -> APIVersion:  # noqa: D102
        raise NotImplementedError()

    @property
    def starting_tip(self) -> Optional[Well]:  # noqa: D102
        raise NotImplementedError()

    @starting_tip.setter
    def starting_tip(self, location: Optional[Well]) -> None:
        raise NotImplementedError()

    def reset_tipracks(self) -> None:  # noqa: D102
        raise NotImplementedError()

    @property
    def default_speed(self) -> float:  # noqa: D102
        raise NotImplementedError()

    @default_speed.setter
    def default_speed(self, speed: float) -> None:
        raise NotImplementedError()

    def aspirate(  # noqa: D102
        self,
        volume: Optional[float] = None,
        location: Union[types.Location, Well] = None,
        rate: float = 1.0,
    ) -> PipetteContext:

        if volume is None or volume == 0:
            # todo(mm, 2021-04-14): If None or 0, use highest volume possible.
            raise NotImplementedError("volume must be specified.")

        if rate != 1:
            raise NotImplementedError(
                "Protocol Engine does not yet support adjusting flow rates."
            )

        if isinstance(location, Well):
            self._engine_client.aspirate(
                pipette_id=self._pipette_id,
                labware_id=location.parent.labware_id,
                well_name=location.well_name,
                well_location=WellLocation(
                    origin=WellOrigin.BOTTOM,
                    # todo(mm, 2021-04-14): Get default offset in well via
                    # self.well_bottom_clearance.aspirate, instead of hard-coding.
                    offset=(0, 0, 1),
                ),
                volume=volume,
            )
        else:
            # todo(mm, 2021-04-14):
            #   * If location is None, use current location.
            #   * If location is a Location (possibly deck coords, or possibly
            #     something like well.top()), use that.
            raise NotImplementedError(
                "locations other than Wells are currently unsupported."
            )

        return self

    def dispense(  # noqa: D102
        self,
        volume: Optional[float] = None,
        location: Union[types.Location, Well] = None,
        rate: float = 1.0,
    ) -> PipetteContext:

        if rate != 1:
            raise NotImplementedError("Flow rate adjustment not yet supported in PE.")

        if volume is None or volume == 0:
            raise NotImplementedError("Volume tracking not yet supported in PE.")

        # TODO (spp:
        #  - Disambiguate location. Cases in point:
        #       1. location not specified; use current labware & well
        #       2. specified location is a Point)
        #  - Use well_bottom_clearance as offset for well_location(?)
        if isinstance(location, Well):
            self._engine_client.dispense(
                pipette_id=self._pipette_id,
                labware_id=location.parent.labware_id,
                well_name=location.well_name,
                well_location=WellLocation(origin=WellOrigin.BOTTOM, offset=(0, 0, 1)),
                volume=volume,
            )
        else:
            raise NotImplementedError(
                "Dispensing to a non-well location not yet supported in PE."
            )
        return self

    def mix(  # noqa: D102
        self,
        repetitions: int = 1,
        volume: Optional[float] = None,
        location: Union[types.Location, Well] = None,
        rate: float = 1.0,
    ) -> PipetteContext:
        raise NotImplementedError()

    def blow_out(  # noqa: D102
        self,
        location: Union[types.Location, Well] = None,
    ) -> PipetteContext:
        raise NotImplementedError()

    def touch_tip(  # noqa: D102
        self,
        location: Optional[Well] = None,
        radius: float = 1.0,
        v_offset: float = -1.0,
        speed: float = 60.0,
    ) -> PipetteContext:
        raise NotImplementedError()

    def air_gap(  # noqa: D102
        self,
        volume: Optional[float] = None,
        height: Optional[float] = None,
    ) -> PipetteContext:
        raise NotImplementedError()

    def return_tip(self, home_after: bool = True) -> PipetteContext:  # noqa: D102
        raise NotImplementedError()

    def pick_up_tip(  # noqa: D102
        self,
        location: Union[types.Location, Well] = None,
        presses: Optional[int] = None,
        increment: Optional[float] = None,
    ) -> PipetteContext:
        # TODO(al, 2021-04-12): What about presses and increment? They are not
        #  supported by PE command. They are also not supported by PD protocols
        #  either.
        if presses is not None or increment is not None:
            raise NotImplementedError()
        if isinstance(location, Well):
            self._engine_client.pick_up_tip(
                pipette_id=self._pipette_id,
                labware_id=location.parent.labware_id,
                well_name=location.well_name,
            )
        else:
            # TODO(al, 2021-04-12): Support for picking up next tip in a labware
            #  and in tipracks associated with a pipette
            raise NotImplementedError()

        return self

    def drop_tip(  # noqa: D102
        self,
        location: Union[types.Location, Well] = None,
        home_after: bool = True,
    ) -> PipetteContext:
        # TODO(al, 2021-04-12): What about home_after?
        if not home_after:
            raise NotImplementedError()
        if isinstance(location, Well):
            self._engine_client.drop_tip(
                pipette_id=self._pipette_id,
                labware_id=location.parent.labware_id,
                well_name=location.well_name,
            )
        else:
            # TODO(al, 2021-04-12): Support for dropping tip in trash.
            raise NotImplementedError()

        return self

    def home(self) -> PipetteContext:  # noqa: D102
        raise NotImplementedError()

    def home_plunger(self) -> PipetteContext:  # noqa: D102
        raise NotImplementedError()

    def distribute(  # noqa: D102
        self,
        volume: Union[float, Sequence[float]],
        source: Well,
        dest: List[Well],
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> PipetteContext:
        raise NotImplementedError()

    def consolidate(  # noqa: D102
        self,
        volume: Union[float, Sequence[float]],
        source: List[Well],
        dest: Well,
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> PipetteContext:
        raise NotImplementedError()

    def transfer(  # noqa: D102
        self,
        volume: Union[float, Sequence[float]],
        source: AdvancedLiquidHandling,
        dest: AdvancedLiquidHandling,
        trash: bool = True,
        **kwargs,  # noqa: ANN003
    ) -> PipetteContext:
        raise NotImplementedError()

    def delay(self) -> None:  # noqa: D102
        raise NotImplementedError()

    def move_to(  # noqa: D102
        self,
        location: types.Location,
        force_direct: bool = False,
        minimum_z_height: Optional[float] = None,
        speed: Optional[float] = None,
    ) -> PipetteContext:
        raise NotImplementedError()

    @property
    def mount(self) -> str:  # noqa: D102
        raise NotImplementedError()

    @property
    def speed(self) -> PlungerSpeeds:  # noqa: D102
        raise NotImplementedError()

    @property
    def flow_rate(self) -> FlowRates:  # noqa: D102
        raise NotImplementedError()

    @property
    def type(self) -> str:  # noqa: D102
        raise NotImplementedError()

    @property
    def tip_racks(self) -> List[Labware]:  # noqa: D102
        raise NotImplementedError()

    @tip_racks.setter
    def tip_racks(self, racks: List[Labware]) -> None:
        raise NotImplementedError()

    @property
    def trash_container(self) -> Labware:  # noqa: D102
        raise NotImplementedError()

    @trash_container.setter
    def trash_container(self, trash: Labware) -> None:
        raise NotImplementedError()

    @property
    def name(self) -> str:  # noqa: D102
        raise NotImplementedError()

    @property
    def model(self) -> str:  # noqa: D102
        raise NotImplementedError()

    @property
    def min_volume(self) -> float:  # noqa: D102
        raise NotImplementedError()

    @property
    def max_volume(self) -> float:  # noqa: D102
        raise NotImplementedError()

    @property
    def current_volume(self) -> float:  # noqa: D102
        raise NotImplementedError()

    @property
    def has_tip(self) -> bool:  # noqa: D102
        raise NotImplementedError()

    @property
    def hw_pipette(self) -> PipetteDict:  # noqa: D102
        raise NotImplementedError()

    @property
    def channels(self) -> int:  # noqa: D102
        raise NotImplementedError()

    @property
    def return_height(self) -> float:  # noqa: D102
        raise NotImplementedError()

    @property
    def well_bottom_clearance(self) -> Clearances:  # noqa: D102
        raise NotImplementedError()

    def pair_with(  # noqa: D102
        self, instrument: PipetteContext
    ) -> PairedInstrumentContext:
        raise NotImplementedError()



"""An emulation of the opentrons magnetic module.

The purpose is to provide a fake backend that responds to GCODE commands.
"""

import logging
from typing import Optional
from opentrons.drivers.mag_deck.driver import GCODES
from opentrons.hardware_control.emulation.parser import Parser, Command
from .abstract_emulator import AbstractEmulator

logger = logging.getLogger(__name__)

GCODE_HOME = GCODES['HOME']
GCODE_MOVE = GCODES['MOVE']
GCODE_PROBE_PLATE = GCODES['PROBE_PLATE']
GCODE_GET_PLATE_HEIGHT = GCODES['GET_PLATE_HEIGHT']
GCODE_GET_CURRENT_POSITION = GCODES['GET_CURRENT_POSITION']
GCODE_DEVICE_INFO = GCODES['DEVICE_INFO']
GCODE_PROGRAMMING_MODE = GCODES['PROGRAMMING_MODE']

SERIAL = "magnetic_emulator"
MODEL = "mag_deck_v20"
VERSION = "2.0.0"


class MagDeckEmulator(AbstractEmulator):
    """Magdeck emulator"""

    def __init__(self, parser: Parser) -> None:
        self.height: float = 0
        self.position: float = 0
        self._parser = parser

    def handle(self, line: str) -> Optional[str]:
        """Handle a line"""
        results = (self._handle(c) for c in self._parser.parse(line))
        joined = ' '.join(r for r in results if r)
        return None if not joined else joined

    def _handle(self, command: Command) -> Optional[str]:
        """Handle a command."""
        logger.info(f"Got command {command}")
        if command.gcode == GCODE_HOME:
            self.height = 0
        elif command.gcode == GCODE_MOVE:
            position = command.params['Z']
            assert isinstance(position, float), f"invalid position '{position}'"
            self.position = position
        elif command.gcode == GCODE_PROBE_PLATE:
            self.height = 45
        elif command.gcode == GCODE_GET_PLATE_HEIGHT:
            return f"height:{self.height}"
        elif command.gcode == GCODE_GET_CURRENT_POSITION:
            return f"Z:{self.position}"
        elif command.gcode == GCODE_DEVICE_INFO:
            return f"serial:{SERIAL} model:{MODEL} version:{VERSION}"
        elif command.gcode == GCODE_PROGRAMMING_MODE:
            pass
        return None



"""Base protocol engine types and interfaces."""
from enum import Enum
from dataclasses import dataclass
from pydantic import BaseModel
from typing import Union, Tuple
from typing_extensions import final

from opentrons.types import DeckSlotName


class DeckSlotLocation(BaseModel):
    """Location for labware placed in a single slot."""

    slot: DeckSlotName


LabwareLocation = Union[DeckSlotLocation]
"""Union of all legal labware locations."""


@final
class WellOrigin(str, Enum):
    """Origin of WellLocation offset."""

    TOP = "top"
    BOTTOM = "bottom"


class WellLocation(BaseModel):
    """A relative location in reference to a well's location."""

    origin: WellOrigin = WellOrigin.TOP
    offset: Tuple[float, float, float] = (0, 0, 0)


class DeckLocation(BaseModel):
    """A symbolic reference to a location on the deck.

    Specified as the pipette, labware, and well. A `DeckLocation` may be
    combined with a `WellLocation` to produce an absolute position in deck
    coordinates.
    """

    pipette_id: str
    labware_id: str
    well_name: str


@final
@dataclass(frozen=True)
class Dimensions:
    """Dimensions of an object in deck-space."""

    x: float
    y: float
    z: float


# TODO(mc, 2021-04-16): reconcile with opentrons_shared_data
# shared-data/python/opentrons_shared_data/pipette/dev_types.py
class PipetteName(str, Enum):
    """Pipette load name values."""

    P10_SINGLE = "p10_single"
    P10_MULTI = "p10_multi"
    P20_SINGLE_GEN2 = "p20_single_gen2"
    P20_MULTI_GEN2 = "p20_multi_gen2"
    P50_SINGLE = "p50_single"
    P50_MULTI = "p50_multi"
    P300_SINGLE = "p300_single"
    P300_MULTI = "p300_multi"
    P300_SINGLE_GEN2 = "p300_single_gen2"
    P300_MULTI_GEN2 = "p300_multi_gen2"
    P1000_SINGLE = "p1000_single"
    P1000_SINGLE_GEN2 = "p1000_single_gen2"



from opentrons import types
from opentrons.protocols.context.instrument import \
    AbstractInstrument
from opentrons.protocols.context.protocol_api.protocol_context import \
    ProtocolContextImplementation
from opentrons.protocols.context.simulator.instrument_context import \
    InstrumentContextSimulation


class ProtocolContextSimulation(ProtocolContextImplementation):
    def load_instrument(self,
                        instrument_name: str,
                        mount: types.Mount,
                        replace: bool) -> AbstractInstrument:
        """Create a simulating instrument context."""
        instr = self._instruments[mount]
        if instr and not replace:
            raise RuntimeError(
                f"Instrument already present in {mount.name.lower()} "
                f"mount: {instr.get_instrument_name()}")

        attached = {att_mount: instr.get('name', None)
                    for att_mount, instr
                    in self._hw_manager.hardware.attached_instruments.items()
                    if instr}
        attached[mount] = instrument_name
        self._hw_manager.hardware.cache_instruments(attached)

        new_instr = InstrumentContextSimulation(
            protocol_interface=self,
            pipette_dict=self._hw_manager.hardware.get_attached_instruments()[
                mount
            ],
            mount=mount,
            instrument_name=instrument_name,
        )
        self._instruments[mount] = new_instr
        self._log.info(f"Instrument {new_instr} loaded")
        return new_instr



import asyncio
import logging

from opentrons.drivers.rpi_drivers.types import USBPort
# NOTE: Must import all modules so they actually create the subclasses
from . import update, tempdeck, magdeck, thermocycler, types  # noqa: F401
from .mod_abc import AbstractModule
from ..execution_manager import ExecutionManager
from .types import InterruptCallback


log = logging.getLogger(__name__)

# TODO (lc 05-12-2021) This is pretty gross. We should think
# of a better way to do this.
MODULE_HW_BY_NAME = {
    cls.name(): cls for cls in AbstractModule.__subclasses__()
}


async def build(
        port: str,
        which: str,
        simulating: bool,
        usb_port: USBPort,
        interrupt_callback: InterruptCallback,
        loop: asyncio.AbstractEventLoop,
        execution_manager: ExecutionManager,
        sim_model: str = None) -> AbstractModule:
    return await MODULE_HW_BY_NAME[which].build(
        port,
        usb_port,
        interrupt_callback=interrupt_callback,
        simulating=simulating,
        loop=loop,
        execution_manager=execution_manager,
        sim_model=sim_model
    )



from __future__ import annotations
import asyncio
from asyncio import Queue
from contextlib import contextmanager
import logging
from typing import (
    Any, Callable, Dict, Sequence, overload, Generic, TypeVar,
    cast, TYPE_CHECKING)
from typing_extensions import Literal

from opentrons.commands import types

if TYPE_CHECKING:
    from opentrons.api.dev_types import Message as SessionStateMessage
    from opentrons.api.calibration import Message as CalibrationStateMessage


MODULE_LOG = logging.getLogger(__name__)


UntypedMessage = Dict[str, Any]


_HandledMessages = TypeVar('_HandledMessages')


class Notifications(Generic[_HandledMessages]):
    def __init__(self, topics: Sequence[str], broker: Broker, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.queue: Queue[UntypedMessage] = Queue(loop=self.loop)
        self.snoozed = False
        self._unsubscribe = [
            broker.subscribe(topic, self.on_notify) for topic in topics]

    @contextmanager
    def snooze(self):
        self.snoozed = True
        try:
            yield
        finally:
            self.snoozed = False

    def on_notify(self, message: UntypedMessage):
        if self.snoozed:
            return
        self.queue.put_nowait(message)

    async def __anext__(self) -> _HandledMessages:
        msg = await self.queue.get()
        return cast(_HandledMessages, msg)

    def __aiter__(self):
        return self


class Broker:
    def __init__(self):
        self.subscriptions = {}
        self.logger = MODULE_LOG

    @overload
    def subscribe(
            self,
            topic: Literal['command'],
            handler: Callable[[types.CommandMessage], None]) -> Callable[[], None]: ...

    @overload
    def subscribe(
            self,
            topic: Literal['calibration'],
            handler: Callable[[CalibrationStateMessage], None])\
        -> Callable[[], None]: ...

    @overload
    def subscribe(
            self,
            topic: Literal['session'],
            handler: Callable[[SessionStateMessage], None]) -> Callable[[], None]: ...

    @overload
    def subscribe(
            self,
            topic: str,
            handler: Callable[[UntypedMessage], None]) -> Callable[[], None]: ...

    def subscribe(self, topic, handler):
        if handler in self.subscriptions.setdefault(topic, []):
            return
        self.subscriptions[topic].append(handler)

        def unsubscribe():
            self.subscriptions[topic].remove(handler)

        return unsubscribe

    @overload
    def publish(
            self, topic: Literal['command'], message: types.CommandMessage) -> None: ...

    @overload
    def publish(
            self,
            topic: Literal['calibration'],
            message: CalibrationStateMessage) -> None: ...

    @overload
    def publish(
            self, topic: Literal['session'], message: SessionStateMessage) -> None: ...

    @overload
    def publish(self, topic: str, message: UntypedMessage) -> None: ...

    def publish(self, topic, message):
        [handler(message) for handler in self.subscriptions.get(topic, [])]

    def set_logger(self, logger):
        self.logger = logger



"""Resource providers."""
from __future__ import annotations

from .id_generator import IdGenerator
from .deck_data_provider import DeckDataProvider
from .labware_data_provider import LabwareDataProvider


class ResourceProviders:
    """ResourceProviders container class.

    Wraps various data providers that define procedures to pull and generate
    data for engine setup and command execution.
    """

    _id_generator: IdGenerator
    _labware_data: LabwareDataProvider
    _deck_data: DeckDataProvider

    @classmethod
    def create(cls) -> ResourceProviders:
        """Create a ResourceProviders container and its children."""
        id_generator = IdGenerator()
        labware_data = LabwareDataProvider()
        deck_data = DeckDataProvider(labware_data=labware_data)

        return cls(
            id_generator=id_generator,
            labware_data=labware_data,
            deck_data=deck_data,
        )

    def __init__(
        self,
        id_generator: IdGenerator,
        labware_data: LabwareDataProvider,
        deck_data: DeckDataProvider,
    ) -> None:
        """Initialize a ResourceProviders container."""
        self._id_generator = id_generator
        self._labware_data = labware_data
        self._deck_data = deck_data

    @property
    def id_generator(self) -> IdGenerator:
        """Get the unique ID generator resource."""
        return self._id_generator

    @property
    def labware_data(self) -> LabwareDataProvider:
        """Get the labware data provider resource."""
        return self._labware_data

    @property
    def deck_data(self) -> DeckDataProvider:
        """Get the deck data provider resource."""
        return self._deck_data



import argparse
from typing import Tuple

from opentrons.hardware_control import ThreadManager, API, adapters
from opentrons.drivers.smoothie_drivers.driver_3_0 import SmoothieDriver_3_0_0


def root_argparser(description: str = None):
    parse = argparse.ArgumentParser(description=description)
    parse.add_argument('-p', '--port',
                       help='serial port of the smoothie',
                       default='', type=str)
    return parse


def build_driver(
        port: str = None)\
        -> Tuple[adapters.SynchronousAdapter, SmoothieDriver_3_0_0]:
    hardware = ThreadManager(API.build_hardware_controller, None, port).sync
    driver = hardware._backend._smoothie_driver
    return hardware, driver



"""Protocol engine state module."""

from .state_store import CommandState, StateStore, StateView
from .labware import LabwareState, LabwareData
from .pipettes import PipetteState, PipetteData, HardwarePipette
from .geometry import GeometryState, TipGeometry
from .motion import MotionState, PipetteLocationData

__all__ = [
    "StateStore",
    "StateView",
    "CommandState",
    "LabwareState",
    "PipetteState",
    "GeometryState",
    "MotionState",
    "LabwareData",
    "PipetteData",
    "HardwarePipette",
    "TipGeometry",
    "PipetteLocationData",
]



"""An emulation of the opentrons temperature module.

The purpose is to provide a fake backend that responds to GCODE commands.
"""

import logging
from typing import Optional

from opentrons.drivers.temp_deck.driver import GCODES
from opentrons.hardware_control.emulation import util
from opentrons.hardware_control.emulation.parser import Parser, Command

from .abstract_emulator import AbstractEmulator


logger = logging.getLogger(__name__)

GCODE_GET_TEMP = GCODES['GET_TEMP']
GCODE_SET_TEMP = GCODES['SET_TEMP']
GCODE_DEVICE_INFO = GCODES['DEVICE_INFO']
GCODE_DISENGAGE = GCODES['DISENGAGE']
GCODE_PROGRAMMING_MODE = GCODES['PROGRAMMING_MODE']

SERIAL = "temperature_emulator"
MODEL = "temp_deck_v20"
VERSION = "v2.0.1"


class TempDeckEmulator(AbstractEmulator):
    """TempDeck emulator"""

    def __init__(self, parser: Parser) -> None:
        self.target_temp = util.OptionalValue[float]()
        self.current_temp = 0.0
        self._parser = parser

    def handle(self, line: str) -> Optional[str]:
        """Handle a line"""
        results = (self._handle(c) for c in self._parser.parse(line))
        joined = ' '.join(r for r in results if r)
        return None if not joined else joined

    def _handle(self, command: Command) -> Optional[str]:
        """Handle a command."""
        logger.info(f"Got command {command}")
        if command.gcode == GCODE_GET_TEMP:
            return f"T:{self.target_temp} C:{self.current_temp}"
        elif command.gcode == GCODE_SET_TEMP:
            temperature = command.params['S']
            assert isinstance(temperature, float),\
                f"invalid temperature '{temperature}'"
            self.target_temp.val = temperature
            self.current_temp = self.target_temp.val
        elif command.gcode == GCODE_DISENGAGE:
            self.target_temp.val = None
            self.current_temp = util.TEMPERATURE_ROOM
        elif command.gcode == GCODE_DEVICE_INFO:
            return f"serial:{SERIAL} model:{MODEL} version:{VERSION}"
        elif command.gcode == GCODE_PROGRAMMING_MODE:
            pass
        return None



"""ProtocolEngine clients."""
from .sync_client import SyncClient
from .transports import AbstractSyncTransport, ChildThreadTransport

__all__ = ["SyncClient", "AbstractSyncTransport", "ChildThreadTransport"]



"""Move to well command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel

from .command import CommandImplementation, CommandHandlers
from .pipetting_common import BasePipettingRequest


class MoveToWellRequest(BasePipettingRequest):
    """A request to move a pipette to a specific well."""

    def get_implementation(self) -> MoveToWellImplementation:
        """Get the move to well request's command implementation."""
        return MoveToWellImplementation(self)


class MoveToWellResult(BaseModel):
    """Result data from the execution of a MoveToWellRequest."""

    pass


class MoveToWellImplementation(
    CommandImplementation[MoveToWellRequest, MoveToWellResult]
):
    """Move to well command implementation."""

    async def execute(self, handlers: CommandHandlers) -> MoveToWellResult:
        """Move the requested pipette to the requested well."""
        await handlers.movement.move_to_well(
            pipette_id=self._request.pipetteId,
            labware_id=self._request.labwareId,
            well_name=self._request.wellName,
        )

        return MoveToWellResult()



"""

Gantry Test
This Test objective will test the Gantry at 85% motor current
with Max speed. This will result to a good assembly vs a bad assembly process.


Author: Carlos Fernandez
"""
import logging
from typing import TYPE_CHECKING

from opentrons.hardware_control.adapters import SynchronousAdapter
from opentrons.drivers.smoothie_drivers.driver_3_0 import \
    SmoothieError, DEFAULT_AXES_SPEED

if TYPE_CHECKING:
    from opentrons.drivers.smoothie_drivers.driver_3_0\
        import SmoothieDriver_3_0_0

from . import args_handler


def setup_motor_current(hardware: SynchronousAdapter,
                        driver: 'SmoothieDriver_3_0_0'):
    # only set the current, keeping all other settings at the driver's default
    driver.set_speed(DEFAULT_AXES_SPEED)
    x_current = hardware.config.high_current['X'] * 0.85
    y_current = hardware.config.high_current['Y'] * 0.85
    driver.set_active_current(
        {'X': x_current, 'Y': y_current})


def bowtie_pattern(X_max: float, Y_max: float,
                   driver: 'SmoothieDriver_3_0_0'):
    zero = 10
    offset = 5
    driver.move({'X': zero, 'Y': zero + offset})
    driver.move({'X': zero, 'Y': Y_max})
    driver.move({'X': X_max, 'Y': zero})
    driver.move({'X': X_max, 'Y': Y_max})


def hourglass_pattern(X_max: float, Y_max: float,
                      driver: 'SmoothieDriver_3_0_0'):
    zero = 10
    offset = 5
    driver.move({'X': zero, 'Y': zero + offset})
    driver.move({'X': X_max, 'Y': zero + offset})
    driver.move({'X': zero, 'Y': Y_max})
    driver.move({'X': X_max, 'Y': Y_max})


def test_axis(axis: str, tolerance: float,
              driver: 'SmoothieDriver_3_0_0'):
    retract_amounts = {
        'X': 3,
        'Y': 3,
        'Z': 2,
        'A': 2,
        'B': 2,
        'C': 2
    }
    # it moves RETRACT_MM away from the endstop
    retract = retract_amounts[axis]
    expected_point = driver.homed_position[axis] + retract
    points = [expected_point - tolerance, expected_point + tolerance]
    # expected_point = home_position + retract
    # safe distance from switch?
    driver.push_speed()
    driver.set_speed(8)
    try:
        driver.move({axis: points[0]})
    except SmoothieError:
        raise Exception('Test Failed: Pressing too soon')
    if axis == 'Y':
        if driver.switch_state[axis] is not False:
            raise Exception('Test Failed: Pressing too soon')
        driver.move({axis: points[1]})
        if driver.switch_state[axis] is not True:
            raise Exception('Test Failed: Not hitting switch')
    else:
        try:
            driver.move({axis: points[1]})
            raise Exception('Test Failed: Not hitting switch')
        except SmoothieError:
            pass
    driver.pop_speed()


def run_x_axis(
        cycles: int, x_max: float, y_max: float, tolerance: float,
        hardware: SynchronousAdapter, driver: 'SmoothieDriver_3_0_0'):
    # Test X Axis
    for cycle in range(cycles):
        print("Testing X")
        setup_motor_current(hardware, driver)
        hourglass_pattern(x_max, y_max, driver)
        try:
            test_axis('X', tolerance, driver)
        except Exception as e:
            print("FAIL: {}".format(e))


def run_y_axis(cycles: int, x_max: float, y_max: float, tolerance: float,
               hardware: SynchronousAdapter, driver: 'SmoothieDriver_3_0_0'):
    # Test Y Axis
    for cycle in range(cycles):
        print("Testing Y")
        setup_motor_current(hardware, driver)
        bowtie_pattern(x_max, y_max, driver)
        try:
            test_axis('Y', tolerance, driver)
        except Exception as e:
            print("FAIL: {}".format(e))
        finally:
            driver.home('Y')  # we tell it to home Y manually


def _exit_test(driver: 'SmoothieDriver_3_0_0'):
    driver._smoothie_reset()
    driver._setup()
    driver.disengage_axis('XYZABC')


if __name__ == '__main__':
    parser = args_handler.root_argparser(
        "run tests checking the assembly of the OT-2's gantry")
    args = parser.parse_args()
    hardware, driver = args_handler.build_driver(args.port)
    num_cycles = 3
    b_x_max = 417.2
    b_y_max = 320
    tolerance_mm = 0.5
    logging.basicConfig(filename='gantry-test.log')
    try:
        hardware.home()
        run_x_axis(num_cycles, b_x_max, b_y_max, tolerance_mm,
                   hardware, driver)
        run_y_axis(num_cycles, b_x_max, b_y_max, tolerance_mm,
                   hardware, driver)
        driver.turn_on_green_button_light()
        print("PASS")
        _exit_test(driver)
    except KeyboardInterrupt:
        print("Test Cancelled")
        driver.turn_on_blue_button_light()
    except Exception as e:
        driver.turn_on_red_button_light()
        print("FAIL: {}".format(e))
        _exit_test(driver)
